#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""The command-line interface to the goo.gl client."""

import click

import lnk.cli
import lnk.config
import lnk.errors

import lnk.googl.link
import lnk.googl.info
import lnk.googl.stats
import lnk.googl.history
import lnk.googl.key

lnk_config = lnk.config.Manager('lnk')['settings']
googl_config = lnk.config.Manager('googl')

link_config = googl_config['commands']['link']
info_config = googl_config['commands']['info']
stats_config = googl_config['commands']['stats']
history_config = googl_config['commands']['history']
key_config = googl_config['commands']['key']

display = history_config['settings']['display']
expanded_default = None if display == 'both' else display == 'expanded'

units = history_config['units']
units += ['{0}s'.format(i) for i in units]

@click.group(invoke_without_command=True,
			 no_args_is_help=True,
			 context_settings=dict(ignore_unknown_options=True),
			 help='goo.gl command-line client.')
@click.option('-v',
			  '--verbose',
			  count=True,
			  help='Increments the level of verbosity.')
@click.option('-l',
			  '--level',
			  nargs=1,
			  type=int,
			  help='Controls the level of verbosity.')
@click.argument('args', nargs=-1)
@click.version_option(version=googl_config['version'],
					  message='goo.gl API v%(version)s')
@click.pass_context
def main(context, verbose, level, args):
	"""
	Main entry-point to the goo.gl API from the command-line.

	All command-calls are handled here. Before passing on all command-specific
	arguments to the appropriate command, the verbosity setting is handled here,
	which is of use when catching any exceptions thrown by any commands.
	Moreover, this entry-point takes care of making the 'link' command of the
	service its default command, such that the user can type 'lnk ...' when
	meaning 'lnk googl link ...'.
	"""
	verbosity = lnk.cli.get_verbosity(verbose, level)
	# Could be the name of the command, or the first argument if the command
	# was not specified (meaning the link command is requested)
	if not args or args[0] not in googl_config['commands'].keys():
		name = googl_config['settings']['command']
	else:
		name = args[0]
		args = args[1:] or ['--help']

	# Pick out the function (command)
	which = globals()[name]
	catch = lnk.errors.Catch(verbosity, which.get_help(context), 'goo.gl')
	catch.catch(which.main, args, standalone_mode=False)

@main.command()
@click.option('-c/-n',
			  '--copy/--no-copy',
			  default=lnk_config['copy'],
			  help='Whether or not to copy the first link to the clipboard.')
@click.option('-q/-l',
			  '--quiet/--loud',
			  default=False,
			  help='Whether or not to print the expanded/shortened links.')
@click.option('-e',
			  '--expand',
			  multiple=True,
			  metavar='URL',
			  help='Expand a short url.')
@click.option('-s',
			  '--shorten',
			  multiple=True,
			  metavar='URL',
			  help='Shorten a long url.')
@click.option('--pretty/--plain',
			  default=link_config['settings']['pretty'],
			  help='Whether to show the links in a pretty box or as a plain list.')
@click.argument('urls', nargs=-1)
def link(copy, quiet, expand, shorten, urls, pretty):
	"""Link shortening and expansion."""
	if not urls and not expand and not shorten:
		raise lnk.errors.UsageError('Please supply at least one URL.')
	lnk.googl.link.echo(copy, quiet, expand, shorten + urls, pretty)

@main.command()
@click.option('-o',
			  '--only',
			  nargs=1,
			  multiple=True,
			  type=click.Choice(info_config['sets']),
			  help='Display only this/these set(s) of information.')
@click.option('-h',
			  '--hide',
			  nargs=1,
			  multiple=True,
			  type=click.Choice(info_config['sets']),
			  help='Hide this/these set(s) of information.')
@click.argument('urls', nargs=-1)
def info(only, hide, urls):
	"""Information about links."""
	# Its' horrible to handle the missing parameter when click
	# throws an exception (doesn't make it accessible), so just do it here.
	if not urls:
		raise lnk.errors.UsageError('Please supply at least one URL.')
	lnk.googl.info.echo(only, hide, urls)

@main.command()
@click.option('-o',
   			  '--only',
   		 	  multiple=True,
   			  type=click.Choice(stats_config['sets']),
   			  help='Display only this/these set(s) of statistics.')
@click.option('-h',
			  '--hide',
			  multiple=True,
   			  type=click.Choice(stats_config['sets']),
   			  help='Hide this/these set(s) of statistics.')
@click.option('-l',
			  '--last',
			  nargs=1,
			  multiple=True,
			  type=click.Choice(stats_config['units']),
			  help='Show statistics for this/these timespan(s).')
@click.option('--forever',
			  is_flag=True,
			  help='Show statistics for all timespans (since forever).')
@click.option('-i',
			  '--info/--no-info',
			  default=stats_config['settings']['info'],
			  help='Also display information for each link.')
@click.option('--limit',
			  type=int,
			  default=stats_config['settings']['limit'],
			  help='Limit the number of links shown per time range.')
@click.option('-a',
			  '--all',
			  '--no-limit',
			  is_flag=True,
			  help='Have no limit on the amount of statistics per timespan.')
@click.option('-f/-s',
			  '--full/--short',
			  default=stats_config['settings']['full-countries'],
			  help='Whether to show full or short (abbreviated) country names.')
@click.argument('urls', nargs=-1)
def stats(only, hide, last, forever, limit, no_limit, info, full, urls):
	"""Statistics and metrics for links."""
	if not urls:
		raise lnk.errors.UsageError('Please supply at least one URL.')
	limit = None if no_limit else limit
	lnk.googl.stats.echo(only, hide, last, forever, limit, info, full, urls)

@main.command()
@click.option('-g',
			 '--generate',
			 is_flag=True,
			 help='Initiates the authorization process.')
def key(generate):
	"""Authorization management."""
	lnk.googl.key.echo(generate)

@main.command()
@click.option('-l',
			  '--last',
			  nargs=2,
			  multiple=True,
			  type=(int, click.Choice(units)),
			  help='Display history of links from this time range, relative to now.')
@click.option('-r',
			  '--range',
			  '--time-range',
			  nargs=4,
			  multiple=True,
			  type=(int, click.Choice(units), int, click.Choice(units)),
			  help='Display history of links from this time range.')
@click.option('-a',
			  '--forever',
			  is_flag=True,
			  help='Display history of links since forever.')
@click.option('--limit',
			  type=int,
			  default=history_config['settings']['limit'],
			  help='Limit the number of links shown per time range.')
@click.option('-a',
			  '--all',
			  '--no-limit',
			  is_flag=True,
			  help='Have no limit on the number of links per timespan.')
@click.option('-e/-s',
			  '--expanded/--short',
			  default=expanded_default,
			  help='Whether to show expanded or shortened links.')
@click.option('-b',
			  '--both',
			  is_flag=True,
			  help='Whether to show expanded and shortened links.')
@click.option('--pretty/--plain',
			  default=history_config['settings']['pretty'],
			  help='Whether to show the history in a pretty box or as a plain list.')
def history(last, time_range, forever, limit, no_limit, expanded, both, pretty):
	"""Retrieve link history."""
	if not last and not time_range:
		forever = True
	# Default case for both
	if not both and expanded is None:
		both = True
	limit = None if no_limit else limit
	lnk.googl.history.echo(last, time_range, forever, limit, expanded, both, pretty)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy
import ecstasy
import os
import pytest
import requests

from collections import namedtuple

import tests.paths
import lnk.bitly.stats
import lnk.bitly.info
import lnk.config

VERSION = 3
API = 'https://api-ssl.bitly.com/v{0}'.format(VERSION)
with open(os.path.join(tests.paths.TEST_PATH, 'bitly', 'token')) as source:
	ACCESS_TOKEN = source.read()

def request_stats(url, endpoint, timespan):
	response = requests.get('{0}/link/{1}'.format(API, endpoint),
							params=dict(units=timespan.span,
										unit=timespan.unit,
										access_token=ACCESS_TOKEN,
										link=url))

	e = endpoint if endpoint != 'clicks' else 'link_clicks'
	data = response.json()['data'][e]

	return {'timespan': timespan, 'data': data}

@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'stats',
		'info',
		'endpoint',
		'url',
		'timespan_tuples',
		'timespans',
		'forever',
		'first_level',
		'second_level',
		'forever_data',
		'timespans_data',
		'default_timespan'
		])

	endpoint = 'countries'
	url = 'http://bit.ly/1V0HZR7'
	timespan_tuples = [(4, 'month'), (3, 'week')]
	timespans = [lnk.bitly.stats.Stats.Timespan(s, u) for s, u in timespan_tuples]
	first_level = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	second_level = ecstasy.beautify('   <-> {0}: {1}', ecstasy.Color.Yellow)
	forever = lnk.bitly.stats.Stats.Timespan(-1, 'day')
	forever_data = request_stats(url, endpoint, forever)
	timespans_data = [request_stats(url, endpoint, i) for i in timespans]

	default_timespan = lnk.bitly.stats.Stats.Timespan(1, 'month')
	with lnk.config.Manager('bitly', write=True) as manager:
		settings = manager['commands']['stats']['settings']
		old = (settings['span'], settings['unit'])
		settings['span'] = default_timespan.span
		settings['unit'] = default_timespan.unit

	stats = lnk.bitly.stats.Stats(raw=True)
	info = lnk.bitly.info.Info(raw=True)

	def finalize():
		with lnk.config.Manager('bitly', write=True) as manager:
			settings = manager['commands']['stats']['settings']
			settings['span'] = old[0]
			settings['unit'] = old[1]

	request.addfinalizer(finalize)

	return Fixture(stats,
				   info,
				   endpoint,
				   url,
				   timespan_tuples,
				   timespans,
				   forever,
				   first_level,
				   second_level,
				   forever_data,
				   timespans_data,
				   default_timespan)


def test_format_makes_lines_pretty(fixture):
	result = fixture.stats.format(None, 'foo', 'bar', None)
	expected = fixture.second_level.format('foo', 'bar')

	assert result == expected


def test_countries_leaves_countries_short(fixture):
	result = fixture.stats.format('countries', 'DE', 123, False)
	expected = fixture.second_level.format('DE', '123')

	assert result == expected


def test_format_expands_countries(fixture):
	result = fixture.stats.format('countries', 'DE', 123, True)
	expected = fixture.second_level.format('Germany', '123')

	assert result == expected

def test_format_handles_unknown_countries(fixture):
	result = fixture.stats.format('countries', 'None', 123, True)
	expected = fixture.second_level.format('Other', '123')

	assert result == expected


def test_format_title_cases_direct(fixture):
	result = fixture.stats.format(None, 'direct', 666, None)
	expected = fixture.second_level.format('Direct', '666')

	assert result == expected

def test_request_format_is_correct(fixture):
	fixture.stats.queue.put((
		fixture.url,
		'countries',
		fixture.forever,
		{
		'unit': fixture.forever.unit,
		'units': fixture.forever.span,
		'link': fixture.url
		}))
	results = {fixture.endpoint: []}
	fixture.stats.request(results)
	result = results[fixture.endpoint][0]

	assert isinstance(result, dict)
	assert 'timespan' in result
	assert isinstance(result['timespan'], fixture.stats.Timespan)
	assert 'data' in result
	assert isinstance(result['data'], list)


def test_requests_countries_well(fixture):
	fixture.stats.queue.put((
		fixture.url,
		'countries',
		fixture.forever,
		{
		'unit': fixture.forever.unit,
		'units': fixture.forever.span,
		'link': fixture.url
		}))
	result = {fixture.endpoint: []}
	fixture.stats.request(result)
	result = result[fixture.endpoint][0]

	assert result == fixture.forever_data

def test_requests_referrers_well(fixture):
	fixture.stats.queue.put((
		fixture.url,
		'referrers',
		fixture.forever,
		{
		'unit': fixture.forever.unit,
		'units': fixture.forever.span,
		'link': fixture.url
		}))
	result = {'referrers': []}
	fixture.stats.request(result)
	result = result['referrers'][0]
	expected = request_stats(fixture.url, 'referrers', fixture.forever)

	assert result == expected

def test_requests_clicks_well(fixture):
	fixture.stats.queue.put((
		fixture.url,
		'clicks',
		fixture.forever,
		{
		'unit': fixture.forever.unit,
		'units': fixture.forever.span,
		'link': fixture.url
		}))
	result = {'clicks': []}
	fixture.stats.request(result)
	result = result['clicks'][0]
	expected = request_stats(fixture.url, 'clicks', fixture.forever)

	assert result == expected

def test_requests_timespan_well(fixture):
	fixture.stats.queue.put((
		fixture.url,
		fixture.endpoint,
		fixture.timespans[0],
		{
		'unit': fixture.timespans[0].unit,
		'units': fixture.timespans[0].span,
		'link': fixture.url
		}))
	results = {fixture.endpoint: []}
	fixture.stats.request(results)

	assert results[fixture.endpoint][0] == fixture.timespans_data[0]


def test_listify_sets_None_if_no_items(fixture):
	data = copy.deepcopy(fixture.forever_data)
	data['data'] = []
	result = fixture.stats.listify(None, [data], False)
	header = fixture.first_level.format('Since forever')
	expected = ['{0}: None'.format(header)]

	assert result == expected


def test_listify_handles_clicks_well(fixture):
	data = copy.deepcopy(fixture.forever_data)
	data['data'] = 123
	result = fixture.stats.listify(None, [data], False)
	header = fixture.first_level.format('Since forever')
	expected = ['{0}: 123'.format(header)]

	assert result == expected


def test_listify_handles_lists_well(fixture):
	data = copy.deepcopy(fixture.forever_data)
	result = fixture.stats.listify(None, [data], False)
	header = fixture.first_level.format('Since forever')
	expected = ['{0}:'.format(header)]
	if fixture.forever_data['data']:
		for i in fixture.forever_data['data']:
			# A way of not having to pop the clicks
			clicks = i['clicks']
			keys = list(i.keys())
			keys.remove('clicks')
			assert len(keys) == 1
			key = i[keys[0]]
			line = fixture.second_level.format(key, clicks)
			expected.append(line)
	else:
		expected[-1] += ' None'

	assert result == expected


def test_get_header_formats_timespans_well(fixture):
	result = fixture.stats.get_header(4, 'weeks')
	expected = fixture.first_level.format('Last 4 weeks:')

	assert result == expected


def test_get_header_handles_forever_well(fixture):
	result = fixture.stats.get_header(-1, None)
	expected = fixture.first_level.format('Since forever:')

	assert result == expected


def test_get_header_removes_unit_if_1(fixture):
	result = fixture.stats.get_header(1, 'day')
	expected = fixture.first_level.format('Last day:')

	assert result == expected


def test_get_header_handles_years_well(fixture):
	result = fixture.stats.get_header(12, 'months')
	expected = fixture.first_level.format('Last year:')

	assert result == expected


def test_get_header_adds_plural_s_if_multiple_years(fixture):
	result = fixture.stats.get_header(24, 'months')
	expected = fixture.first_level.format('Last 2 years:')

	assert result == expected


def test_lineify_formats_headers_well(fixture):

	data = {fixture.endpoint: [copy.deepcopy(fixture.forever_data)]}
	result = fixture.stats.lineify(data, False)

	assert 'Countries:' in result


def test_get_timespans_removes_duplicates(fixture):
	timespans = list(fixture.timespan_tuples)
	timespans += [fixture.timespan_tuples[0]] * 10
	result = fixture.stats.get_timespans(timespans, False)

	assert result == set(fixture.timespans)


def test_get_timespans_picks_default_timespan_if_no_times(fixture):
	result = fixture.stats.get_timespans([], False)

	assert result == set([fixture.default_timespan])


def test_get_timespans_handles_forever_well(fixture):
	result = fixture.stats.get_timespans([], True)

	assert result == set([fixture.forever])


def test_get_timespans_handles_year_well(fixture):
	result = fixture.stats.get_timespans(((1, 'year'),), False)
	expected = set([fixture.stats.Timespan(12, 'months')])

	assert result == expected


def test_get_timespans_works(fixture):
	result = fixture.stats.get_timespans(fixture.timespan_tuples, False)

	assert result == set(fixture.timespans)


def test_get_stats_works_for_single_endpoint(fixture):
	result = fixture.stats.get_stats(fixture.url,
									   [fixture.forever],
									   [fixture.endpoint])
	expected = {fixture.endpoint: [fixture.forever_data]}

	assert result == expected

def test_get_stats_works_for_many_timespans(fixture):
	result = fixture.stats.get_stats(fixture.url,
									   fixture.timespans,
									   [fixture.endpoint])

	assert isinstance(result, dict)
	assert fixture.endpoint in result
	assert isinstance(result[fixture.endpoint], list)

	assert all(i in fixture.timespans_data for i in result[fixture.endpoint])


def test_get_stats_handles_plural_s_in_timespan_well(fixture):
	timespans = copy.deepcopy(fixture.timespans)
	timespans[0] = fixture.stats.Timespan(timespans[0].span,
										  timespans[0].unit + 's')
	result = fixture.stats.get_stats(fixture.url,
								  	   timespans,
								       [fixture.endpoint])
	expected = copy.deepcopy(fixture.timespans_data)
	expected[0]['timespan'] = timespans[0]

	assert isinstance(result, dict)
	assert fixture.endpoint in result

	result = result[fixture.endpoint]

	assert isinstance(result, list)

	for i, j in zip(expected, result):
		i['data'].sort()
		j['data'].sort()

	assert all(i in expected for i in result)


def test_get_stats_works_for_many_endpoints(fixture):
	endpoints = ['clicks', 'referrers', 'countries']
	result = fixture.stats.get_stats(fixture.url,
									   [fixture.forever],
									   endpoints)

	expected = {fixture.endpoint: [fixture.forever_data]}
	endpoints.remove(fixture.endpoint)
	for endpoint in endpoints:
		data = request_stats(fixture.url, endpoint, fixture.forever)
		expected[endpoint] = [data]

	assert result == expected


def test_fetch_works_for_single_url(fixture):
	result = fixture.stats.fetch([fixture.endpoint],
								 [],
								 fixture.timespan_tuples,
								 False,
								 None,
								 False,
								 False,
								 [fixture.url])
	expected = ['URL: {0}'.format(fixture.url)]
	data = copy.deepcopy(fixture.timespans_data)
	data = {fixture.endpoint: data}
	expected += fixture.stats.lineify(data, False)

	assert len(result) == 1
	assert sorted(result[0]) == sorted(expected)


def test_fetch_limits_well(fixture):
	result = fixture.stats.fetch([fixture.endpoint],
								 [],
								 [],
								 True,
								 1,
								 False,
								 False,
								 [fixture.url])
	expected = ['URL: {0}'.format(fixture.url)]
	data = copy.deepcopy(fixture.forever_data)
	while len(data['data']) > 1:
		data['data'].pop()
	assert len(data['data']) == 1
	data = {fixture.endpoint: [data]}
	expected += fixture.stats.lineify(data, False)

	assert len(result) <= 1
	assert sorted(result[0]) == sorted(expected)

def test_fetch_works_for_many_urls(fixture):
	urls = [fixture.url, 'http://bit.ly/1Km6CB1']
	result = fixture.stats.fetch([fixture.endpoint],
								 [],
								 [],
								 True,
								 None,
								 False,
								 False,
								 urls)
	first = ['URL: {0}'.format(fixture.url)]
	data = {fixture.endpoint: [copy.deepcopy(fixture.forever_data)]}
	first += fixture.stats.lineify(data, False)

	second = ['URL: {0}'.format(urls[1])]
	data = request_stats(urls[1], fixture.endpoint, fixture.forever)
	data = {fixture.endpoint: [data]}
	second += fixture.stats.lineify(data, False)

	expected = [sorted(first), sorted(second)]

	assert [sorted(i) for i in result] == expected


def test_fetch_adds_info_well(fixture):
	result = fixture.stats.fetch([fixture.endpoint],
								 [],
								 fixture.timespan_tuples,
								 False,
								 None,
								 True,
								 False,
								 [fixture.url])
	expected = fixture.info.fetch([], [], False, [fixture.url])[0]
	data = {fixture.endpoint: fixture.timespans_data}
	expected += fixture.stats.lineify(data, False)

	assert len(result) == 1
	assert sorted(result[0]) == sorted(expected)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path
import pytest
import re
import sys

from collections import namedtuple

import tests.paths
import lnk.bitly.key
import lnk.config
import lnk.errors

@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'key',
		'login',
		'password'
		])

	with lnk.config.Manager('bitly', write=True) as config:
		old_key = config['key']
		old_login = config['login']

	def finalize():
		with lnk.config.Manager('bitly', write=True) as config:
				config['key'] = old_key
				config['login'] = old_login

	request.addfinalizer(finalize)

	key = lnk.bitly.key.Key(raw=True)
	login = 'lnktest'
	password = 'ilovecats'

	return Fixture(key, login, password)


def test_key_generation(fixture):
	try:
		key = fixture.key.fetch(None,
								fixture.login,
								fixture.password,
								True,
								False)

		assert isinstance(key, str)
		assert len(key) == 40
		assert re.match(r'[a-z\d]', key)

		here = os.path.dirname(os.path.abspath(__file__))
		path = os.path.join(here, 'token')
		with open(path, 'wt') as destination:
			destination.write(key)
	except lnk.errors.HTTPError:
		_, error, _ = sys.exc_info()
		if 'RATE_LIMIT_EXCEEDED' not in error.level(1)[1]:
			raise error


def test_key_is_hidden_if_show_false(fixture):
	try:
		result = fixture.key.fetch(None,
								  fixture.login,
								  fixture.password,
								  False,
								  False)

		assert isinstance(result, str)
		assert result == ''
	except lnk.errors.HTTPError:
		_, error, _ = sys.exc_info()
		if 'RATE_LIMIT_EXCEEDED' not in error.level(1)[1]:
			raise error


def test_key_generation_fails_for_bad_login(fixture):
	with pytest.raises(lnk.errors.HTTPError):
		fixture.key.fetch(None, 'bad_login', fixture.password, False, False)


def test_key_generation_fails_for_bad_password(fixture):
	with pytest.raises(lnk.errors.HTTPError):
		fixture.key.fetch(None, fixture.login, 'bad_password', False, False)


def test_who_works_if_someone_is_logged_in(fixture):
	with lnk.config.Manager('bitly', write=True) as config:
		config['login'] = fixture.login
	result = fixture.key.fetch(None, None, None, False, True)
	expected = '{0}\n'.format(fixture.login)

	assert result == expected


def test_who_works_if_nobody_is_logged_in(fixture):
	with lnk.config.Manager('bitly', write=True) as config:
		config['login'] = None
	result = fixture.key.fetch(None, None, None, False, True)
	expected = 'Nobody.\n'

	assert result == expected



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Information-retrieval for goo.gl link."""

from __future__ import unicode_literals

import click

from datetime import datetime

import lnk.abstract
import lnk.beauty

from lnk.googl.command import Command

def echo(*args):
	"""
	Executes an info command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to an
						 Info instance's fetch() method.
	"""
	click.echo(Info().fetch(*args))

class Info(Command):
	"""
	Class to retrieve information about a goo.gl link.

	Information about a link may include when it was created, its expanded
	url and its status. This information is returned in a pretty list. The
	list can be put into a box for terminal output, or can be returned raw
	for internal use (such as by the stats command).

	Note:
		A 'goo.gl link' is a link shortened with bit.ly.

	Attributes:
		raw (bool): Whether or not to return the output as a raw list,
					or as a string, in a box.
		sets (dict): A complete dictionary of the available sets of information.
		reverse (dict): A reverse mapping of the above-mentioned sets.
	"""
	def __init__(self, raw=False):
		super(Info, self).__init__('info')

		self.raw = raw
		self.sets = self.config['sets']
		# Dictionary comprehensions not available for Python < 2.7
		self.reverse = dict((value, key) for key, value in self.sets.items())

	def fetch(self, only, hide, urls):
		"""
		Fetches the link information.

		Arguments:
			only (tuple): A tuple of strings representing the sets to include
						  in the response ('only' these will be included).
			hide (tuple): A tuple of strings representing the sets to hide
						  from the response (either from all possible sets
						  if only is empty, or else from those selected).
			urls (tuple): A tuple of urls to fetch information for.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is True,
			else a boxified, pretty string.
		"""
		sets = lnk.abstract.filter_sets(self.sets, only, hide)
		result = []
		threads = []
		for url in urls:
			self.queue.put(url.strip())
			t = self.new_thread(self.get_info, sets.values(), result)
			threads.append(t)
		self.join(threads)

		return result if self.raw else lnk.beauty.boxify(result)

	def get_info(self, sets, result):
		"""
		Requests all information about a url.

		This method is run in a separate thread for each URL, and as such
		uses a queue to get a URL and then appends the information to the
		list supplied in the argument. This is done in a thread-safe, locked
		way.

		Arguments:
			sets (dict): The sets of information to include in the data.
			result (list): The list of results for each URL, to which to
						   append the data requested.
		"""
		url = self.queue.get()
		data = self.request(url)
		selection = dict((key, data[key]) for key in data if key in sets)
		lines = self.lineify(url, selection)
		with self.lock:
			result.append(lines)

	def request(self, url):
		"""
		Requests information for a url.

		Arguments:
			url (str): The goo.gl link to request information for.

		Returns:
			The information for the link supplied.
		"""
		what = "get information for '{0}'".format(url)
		response = self.get(url, 'FULL', what)

		return response

	def lineify(self, url, data):
		"""
		Returns a list of lines for given (information) data.

		The lines returned include a header with the relevant URL,
		followed by string-formatted key/value pairs.

		Arguments:
			url (str): The URL of which the data should be formatted.
			data (dict): The data (information) for the url.
		Returns:
			A list of lines, ready for output.
		"""
		lines = ['URL: {0}'.format(url)]
		for key, value in data.items():
			lines.append(self.format(key, value))

		return lines

	def format(self, key, value):
		"""
		Handles formatting of a data-point.

		Handles parsing of the ISO-8601-formatted timestamp from Google's API
		into a proper string-representation (ctime). Also formats the key-value
		pair.

		Arguments:
			key (str): The key of the data-point.
			value (str): The value of the data-point.

		Returns:
			A formatted string of the schema '<key>: <value>'.
		"""
		key = self.reverse[key]
		if key == 'created':
			# Ignore the irrelevant part of the ISO format
			relevant = value[:value.find('.')]
			parsed = datetime.strptime(relevant, '%Y-%m-%dT%H:%M:%S')
			value = parsed.ctime()

		return '{0}: {1}'.format(key.title(), value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import pytest
import requests
import threading

from collections import namedtuple

import tests.paths

import lnk.errors
import lnk.abstract

class Command(lnk.abstract.AbstractCommand):
	def __init__(self):
		super(Command, self).__init__('test', 'do')


@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'command',
		'url',
		'version',
		'api',
		'endpoints',
		'settings',
		])

	directory = os.path.dirname(os.path.abspath(__file__))
	path = os.path.join(tests.paths.CONFIG_PATH, 'test.json')
	with open(os.path.join(directory, 'test.json'), 'rt') as source:
		config = json.load(source)
		with open(path, 'wt') as destination:
			json.dump(config, destination)

	def finalize():
		os.remove(path)

	request.addfinalizer(finalize)

	return Fixture(Command(),
				   config['url'],
				   config['version'],
				   '{0}/v{1}'.format(config['url'], config['version']),
				   config['commands']['do']['endpoints'],
				   config['commands']['do']['settings'])


def test_has_all_attributes(fixture):
	assert hasattr(fixture.command, 'url')
	assert hasattr(fixture.command, 'api')
	assert hasattr(fixture.command, 'config')
	assert hasattr(fixture.command, 'endpoints')
	assert hasattr(fixture.command, 'settings')


def test_attributes_are_correctly_set(fixture):
	assert fixture.command.url == fixture.url
	assert fixture.command.api == fixture.api
	assert fixture.command.endpoints == fixture.endpoints
	assert fixture.command.settings == fixture.settings


def test_default_GET_works_well(fixture):
	endpoint = fixture.endpoints[0]
	url = '{0}/{1}'.format(fixture.api, endpoint)

	response = fixture.command.get(endpoint)

	assert isinstance(response, requests.Response)
	assert response.request.method == 'GET'
	assert response.url == url

	response = fixture.command.get(endpoint,
								   parameters=dict(cats='awesome'))

	assert isinstance(response, requests.Response)
	assert response.request.method == 'GET'
	assert response.url == '{0}?cats=awesome'.format(url)


def test_default_POST_works_well(fixture):
	endpoint = fixture.endpoints[1]
	url = '{0}/{1}'.format(fixture.url, endpoint)

	response = fixture.command.post(endpoint)

	assert isinstance(response, requests.Response)
	assert response.request.method == 'POST'
	assert response.url == url

	response = fixture.command.post(endpoint,
									data=dict(cats='awesome'))

	assert isinstance(response, requests.Response)
	assert response.request.method == 'POST'

	assert response.request.body == 'cats=awesome'


def test_new_thread_sets_up_thread_well(fixture):
	def endless():
		while True:
			pass
	thread = fixture.command.new_thread(endless)

	assert isinstance(thread, threading.Thread)
	assert thread.is_alive()
	assert thread.daemon

	thread.join(0.1)


def test_exception_handling_for_threads_works(fixture):
	def throws():
		raise RuntimeError('Meh')
	thread = fixture.command.new_thread(throws)

	assert isinstance(thread, threading.Thread)
	
	thread.join()

	assert fixture.command.error is not None
	assert isinstance(fixture.command.error, RuntimeError)
	assert str(fixture.command.error) == 'Meh'

	fixture.command.error = None


def test_join_method_works(fixture):
	threads = [fixture.command.new_thread(lambda: 1) for _ in range(10)]
	fixture.command.join(threads)

	assert not any(thread.is_alive() for thread in threads)


def test_join_method_throws_if_unjoinable(fixture):
	def blocks():
		while True:
			pass
	thread = fixture.command.new_thread(blocks)

	with pytest.raises(lnk.errors.InternalError):
		fixture.command.join([thread], timeout=0.1)


def test_join_method_re_raises_thread_exceptions(fixture):
	def throws():
		raise RuntimeError
	thread = fixture.command.new_thread(throws)

	with pytest.raises(RuntimeError):
		fixture.command.join([thread])


def test_fetch_not_implemented(fixture):
	with pytest.raises(NotImplementedError):
		fixture.command.fetch()


def test_filter_sets_filters_well(fixture):
	base = dict((i, None) for i in 'abcde')
	only = ['a', 'c', 'e']
	result = lnk.abstract.filter_sets(base, only, [])

	assert sorted(result.keys()) == only


def test_filter_sets_hides_well(fixture):
	base = dict((i, None) for i in 'abcde')
	hide = ['a', 'c', 'e']
	result = lnk.abstract.filter_sets(base, [], hide)

	assert sorted(result.keys()) == ['b', 'd']



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

import os
import json
import pytest

from collections import namedtuple

import tests.paths

from lnk.config import configure

@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'service',
		'command',
		'path',
		'original',
		'changed',
		'keys',
		'old_values',
		'new_values'
		])
	service = 'test_command'
	command = 'cmd'

	here = os.path.dirname(os.path.abspath(__file__))
	original_file = '{0}.json'.format(service)
	path = os.path.join(tests.paths.CONFIG_PATH, original_file)
	with open(os.path.join(here, original_file)) as source:
		original = json.load(source)
		with open(path, 'w') as test:
			json.dump(original, test)

	changed_file = '{0}.json'.format('changed_command')
	with open(os.path.join(here, changed_file)) as source:
			changed = json.load(source)

	keys = ['rock', 'water']
	old_values = ['solid', 'liquid']
	new_values = ['gas', 'solid']

	def finalize():
		os.remove(path)

	request.addfinalizer(finalize)

	return Fixture(service,
				   command,
				   path,
				   original,
				   changed,
				   keys,
				   old_values,
				   new_values)

def test_configure_can_be_quiet(fixture):
	result = configure.configure(fixture.service,
								None,
								fixture.keys,
								fixture.old_values,
								True,
								False)
	assert result is None


def test_configure_updates_well(fixture):
	configure.configure(fixture.service,
						None,
						fixture.keys,
						fixture.new_values,
						True,
						False)

	with open(fixture.path) as source:
		result = json.load(source)

	assert result == fixture.changed


def test_configure_returns_correct_output(fixture):
	result = configure.configure(fixture.service,
								 None,
								 fixture.keys,
								 fixture.old_values,
								 False,
								 False)
	lines = []
	# Note that we updated back to the old values
	# (that's why the name switch)
	items = zip(fixture.keys, fixture.new_values, fixture.old_values)
	for key, old, new in items:
		line = '{0}: {1} => {2}'.format(key, old, new)
		lines.append(line)
	expected = '{0}\n'.format('\n'.join(lines))

	assert result == expected


def test_configure_updates_command_settings_well(fixture):
	configure.configure(fixture.service,
						fixture.command,
						['foo', 'bar'],
						[False, 123],
						True,
						False)
	with open(fixture.path) as source:
		result = json.load(source)

	expected = fixture.original
	settings = expected['commands']['cmd']['settings']
	settings['foo'] = False
	settings['bar'] = 123

	assert result == expected


def test_format_value_formats_integers():
	result = configure.format_value(5)

	assert result == '5'


def test_format_value_formats_lists():
	items = [1, 2, 3, 4]
	result = configure.format_value(items)
	expected = ', '.join(map(str, items))

	assert result == expected


def test_format_value_formats_dictionaries():
	items = {'foo': 1, 'bar': True, 'cats': 'please'}
	result = configure.format_value(items)
	expected = ', '.join('{0}: {1}'.format(k, v) for k, v in items.items())

	assert result == expected


def test_update_formats_well_if_no_value_given():
	result = configure.update({'foo': 0}, 'foo', None)
	expected = 'foo: 0'

	assert result == expected


def test_update_formats_well_if_value_given():
	result = configure.update({'foo': 0}, 'foo', [1])
	expected = 'foo: 0 => 1'

	assert result == expected


def test_update_updates_well():
	settings = {'foo': 0}
	configure.update(settings, 'foo', [1])
	
	assert settings['foo'] == 1



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Includes the base-class for all commands of any service."""

from __future__ import unicode_literals

import ecstasy
import requests
import threading
import sys

try:
	from Queue import Queue
except ImportError:
	from queue import Queue

import lnk.config
import lnk.errors

class AbstractCommand(object):
	"""
	Abstract-base-class for all commands of any service.

	This class' constructor handles the bulk of configuration every
	command needs, such as fetching the service's API url, available
	endpoints and default settings. It also gives each command a queue
	and lock for threading, as well as a few other things most, if not
	all, commands need. The class defines an interface all commands must
	have, with some of AbstractCommand's methods, such as fetch(), not
	being implemented and throwing a NotImplementedError if called directly.
	An AbstractCommand must have knowledge about the service that the class
	subclassing it uses (e.g. bit.ly or tinyurl), as well as about the
	name of the command (e.g. 'link' or 'stats').

	Attributes:
		url (str): The URL of the API.
		api (str): The URL of the API, joined with its version. Endpoints can
				   be joined to this string to form a full URL (without
				   parameters) for a request.
		config (dict): The configuration data of a command.
		endpoints (dict): The endpoints for a command.
		settings (dict): The default settings of a command.
		sets (dict|None): If available, the data sets/categories that the
						  command allows, else None if the command has no
						  such thing (e.g. the 'link' command).
		queue (Queue.Queue): A queue for thread-safe data-passing.
		lock (threading.Lock): A lock object for thread-safe actions.
		error (Exception): The last exception thrown by a thread started
						   with the new_thread method. This is useful to
						   see if a thread threw an exception which would
						   otherwise not be properly handled (because you
						   can't catch exceptions from a child-thread in the
						   main thread).
		parameters (dict): Dictionary for the parameters of an HTTP request.
		list_item (str): A string, formatted with ecstasy, that should be used
						 to format a list-item (e.g. for the stats command).
						 It already includes the necessary markup such that
						 str.format() can be used on it directly with the
						 string to be formatted.
	"""
	def __init__(self, service, command):
		with lnk.config.Manager(service) as manager:
			self.url = manager['url']
			self.api = '{0}/v{1}'.format(self.url, manager['version'])
			self.config = manager['commands'][command]
			self.endpoints = self.config['endpoints']
			self.settings = self.config.get('settings')
			self.sets = self.config.get('sets')
		self.queue = Queue()
		self.lock = threading.Lock()
		self.error = None
		self.parameters = {}
		self.list_item = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)

	def fetch(self, *args):
		"""
		Abstract method to fetch command-specific data.

		Arguments:
			args (variadic): Whatever arguments the overriden method takes.

		Raises:
			NotImplementedError: When called directly.
		"""
		raise NotImplementedError

	def get(self, endpoint, parameters=None):
		"""
		Base method to perform an HTTP request with the GET method.

		Arguments:
			endpoint (str): The endpoint at which to request data.
			parameters (dict): Additional parameters to pass with the request.

		Return:
			The requests.Response object resulting from the request.
		"""
		url = '{0}/{1}'.format(self.api, endpoint)
		if not parameters:
			parameters = self.parameters
		else:
			parameters.update(self.parameters)

		return requests.get(url, params=parameters, timeout=60)

	def post(self, endpoint, authorization=None, data=None):
		"""
		Base method to perform an HTTP request with the POST method.

		Arguments:
			endpoint (str): The endpoint at which to send data.
			authorization (tuple): Optionally, a (login, password) tuple that
								   should be used for HTTP authorization.
			data (dict): Optionally, data to send with the request.

		Return:
			The requests.Response object resulting from the request.
		"""
		url = '{0}/{1}'.format(self.url, endpoint)

		return requests.post(url, auth=authorization, data=data, timeout=60)

	def new_thread(self, function, *args, **kwargs):
		"""
		Runs a function in a new thread and returns the thread.

		The function is run in a new thread in way that all positional
		and keyword arguments can be forwarded to the function. Additionally,
		extra measures are taken to wrap the function into another proxy
		function that handles exceptions which would otherwise not be handled.
		If the function to be called throws an exception, this is recorded
		and the exception is assigned to 'error' attribute, where it can later
		be checked.

		Arguments:
			function (func): The function to execute.
			args (variadic): The positional arguments to pass to the function
							 when calling it.
			kwargs (variadic): The keyword arguments to pass to the function
							   when calling it.

		Returns:
			The started (!) thread in which the function is being executed.
		"""
		def proxy(*args, **kwargs):
			"""Proxy function for concurrent exception-handling."""
			try:
				function(*args, **kwargs)
			except Exception:
				_, self.error, _ = sys.exc_info()
		thread = threading.Thread(target=proxy, args=args, kwargs=kwargs)
		thread.daemon = True
		thread.start()

		return thread

	def join(self, threads, timeout=120):
		"""
		Joins a list of thread and checks for errors.

		Each thread is join with a timeout period as specified by the timeout
		parameter of this function. If an error was found in the 'error'
		attribute, it is re-raised so that it can be caught in the main thread.

		Arguments:
			threads (list): The list of threading.Threads to join.
			timeout (float): A floating-point number specifying the number of
							 seconds to wait when joining. Defaults to 60
							 seconds.

		Raises:
			lnk.errors.InternalError: If a thread could not be joined in the given
								  timeout period (i.e. if it is still alive after
								  joining).
			Other errors: If a thread threw an exception, this exception is
						  re-raised in the main thread.
		"""
		for thread in threads:
			thread.join(timeout=timeout)
			if thread.is_alive():
				raise lnk.errors.InternalError('Could not join thread.')
		if self.error:
			raise self.error

def filter_sets(all_sets, only, hide):
	"""
	Filters a set of categories.

	This method is used by many commands that use some sort of dictionary
	of available data sets/categories, which must filter those sets according
	to the ones the user wants to have included in the response by that command.

	Arguments:
		all_sets (dict): The base dictionary of all available sets (from which
						 a subset should be filtered).
		only (tuple): A tuple of the names of the categories/sets to include.
		hide (tuple): A tuple of the names of the categories/sets to exclude.
					  Note that sets are hidden after the 'only' sets have
					  been processed, i.e. if a certain set is in 'only' and
					  in 'hide', it will first be selected and then discared
					  again (which would make no sense). Usually either 'only'
					  or 'hide' is empty.

	Returns:
		A dictionary containing the filtered key/value pairs.
	"""
	filtered = {}
	only = only or all_sets
	for key, value in all_sets.items():
		if key in only and key not in hide:
			filtered[key] = value

	return filtered



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Statistics and metrics retrieval for bitlinks."""

from __future__ import unicode_literals

import click
import ecstasy
import time

from collections import namedtuple

import lnk.abstract
import lnk.beauty
import lnk.bitly.info
import lnk.countries
import lnk.errors

from lnk.bitly.command import Command

def echo(*args):
	"""
	Executes a stats command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 Stats instance's fetch() method.
	"""
	click.echo(Stats().fetch(*args))

class Stats(Command):
	"""
	Class to retrieve statistics and info for one or more bitlinks.

	The statistics for a link include its referrers (i.e. from where the link
	was opened), the countries from where the link was opened and of course the
	number of clicks. These statistics can be retrieved 'since-forever', but
	also for specific (possibly open-ended) time-ranges, such as for 'the last
	5 months' or 'between 4 days and 2 minute ago'. Additionally, these
	statistics can be paired with information about each link, retrieved from
	the 'info' command, thereby making the stats command the ultimate
	destination for link statistics *and* information. Output may, as always,
	be in raw format for internal use or in a pretty box. Multiple URLs are
	fully supported.

	Attributes:
		raw (bool): Whether to return the output in raw format for internal use,
					or in a pretty string-representation for outside-display.
		info (lnk.bitly.info.Info): A lnk.bitly.info.Info instance to retrieve
								addittional information for a bitlink.
	"""

	Timespan = namedtuple('Timespan', ['span', 'unit'])

	def __init__(self, raw=False):
		super(Stats, self).__init__('stats')

		self.raw = raw
		self.info = lnk.bitly.info.Info(raw=True)
		self.parameters['timezone'] = time.tzname[0]

	def fetch(self, only, hide, times, forever, limit, add_info, full, urls):
		"""
		Fetches statistics for one or more URLs.

		Arguments:
			only (tuple): A tuple of strings representing the sets to include
						  in the response ('only' these will be included).
			hide (tuple): A tuple of strings representing the sets to hide
						  from the response (either from all possible sets
						  if only is empty, or else from those selected).
			times (tuple): A tuple of tuples of the schema (<span>, <unit>),
						   representing the timespans for which to fetch
						   statistics.
			forever (bool): Whether to include the statistics 'since forever'.
			limit (int): A limit on the number of items fetched per timespan.
			add_info (bool): Whether or not to add additional information for
							 each link.
			full (bool): Whether to show full country names, or short ISO
						 abbreviations.
			urls (tuple): A tuple of urls to fetch statistics for.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is True,
			else a boxified, pretty string.
		"""
		self.parameters['limit'] = limit

		sets = lnk.abstract.filter_sets(self.sets, only, hide)
		timespans = self.get_timespans(times, forever)
		info = []
		if add_info:
			info = self.info.fetch([], [], False, urls)
			if not info:
				raise lnk.errors.InternalError('Could not fetch additional info.')

		results = []
		for n, url in enumerate(urls):
			header = info[n] if add_info else ['URL: {0}'.format(url)]
			data = self.get_stats(url, timespans, sets)
			lines = self.lineify(data, full)

			results.append(header + lines)

		return results if self.raw else lnk.beauty.boxify(results)

	def get_stats(self, url, timespans, sets):
		"""
		Retrieves the statistics for a single url.

		The statistics returned are for all timespans supplied, filtered
		according to the sets of statistics wanted.

		Note:
			This method works with threads. For each category and each
			timespan, a new request must be made. Each request is made
			in a separate thread.

		Arguments:
			url (str): The relevant URL to fetch statistics for.
			timespans (tuple): A tuple of tuples of the schema (<span>, <unit>),
							   representing the timespans for which to fetch
							   statistics.
			sets (tuple): The sets of statistics wanted in the response
						 (others are discarded).

		Returns:
			A single 'result', which is a dictionary where the keys are
			the sets (categories) and the values a list of the statistics
			that were fetched for each timespan. Each item in this list
			is in turn a dictionary with one key/value pair for the timespan
			and another for the data retrieved in that category,
			for that timespan. This data is either an integer for the 'clicks'
			category, or a list representing each individual data-point that
			could be fetched. Each of those data-points has a 'clicks' key
			whose value is an integer representing the clicks for that
			data-point. A data-point also has another key, which depends on the
			category, e.g. 'country' for the 'countries' category.
		"""
		parameters = {'link': url}
		result = {}
		threads = []
		for endpoint in sets:
			result[endpoint] = []
			for timespan in timespans:

				parameters['unit'] = timespan.unit

				if timespan.unit.endswith('s'):
					# Get rid of the plural s in e.g. 'weeks'
					parameters['unit'] = timespan.unit[:-1]

				parameters['units'] = timespan.span

				self.queue.put((url, endpoint, timespan, parameters))
				threads.append(self.new_thread(self.request, result))
		self.join(threads)

		return result

	def request(self, results):
		"""
		Requests statistics for a given configuration.

		The URL, endpoint (set/category) and timespan are all fetched from
		a queue, because this method is always run in a separate thread for
		each configuration of the above parameters. The dictionary to which
		to insert the data must be passed as an argument. The data is
		inserted in a thread-safe, locked way.

		Arguments:
			results (dict): The results dictionary for the URL, to which to
							insert the retrieved statistics.
		"""
		url, endpoint, timespan, parameters = self.queue.get()

		response = self.get(self.endpoints[endpoint], parameters)
		what = "retrieve {0} for '{1}'".format(endpoint, url)
		response = self.verify(response, what)

		# For 'clicks' the key has a different name than the endpoint
		e = endpoint if endpoint != 'clicks' else 'link_clicks'
		data = {'timespan': timespan, 'data': response[e]}

		with self.lock:
			results[endpoint].append(data)

	def get_timespans(self, times, forever):
		"""
		Parses the timespans passed to fetch and returns Timespan objects.

		The timespans passed to the command-line interface and ultimately
		to the fetch() method are of the schema '(<span>, <unit>)'. These
		are parsed into Stats.Timespan objects. Parsing handles cases such
		as where the unit is 'year'. Bitly cannot handle 'year' as a unit,
		so the timespan must be converted into weeks. Also the 'forever' flag
		is handled here. Moreover, this method retrieves the default timespan
		from the configuration file if no timespans were passed to the method
		(and the command on the CLI).

		Arguments:
			times (tuple): The timespans of the schema '(<span>, <unit>)'.
			forever (bool): Whether to include the 'since forever' timespan.

		Returns:
			A list of Stats.Timespan objects.
		"""
		timespans = set()
		if forever:
			# -1 = since forever (unit could be any)
			timespans.add(Stats.Timespan(-1, 'day'))
		if times:
			for span, unit in times:
				if 'year' in unit:
					span *= 12
					unit = 'months'
				timespans.add(Stats.Timespan(span, unit))
		elif not forever:
			# Get the default from the settings
			unit = self.settings['unit']
			span = self.settings['span']
			if not unit or not span:
				raise lnk.errors.InternalError('Default timespan is invalid!')
			if unit == 'forever':
				timespans.add(Stats.Timespan(-1, 'day'))
			elif 'year' in unit:
				timespans.add(Stats.Timespan(span * 12, 'months'))
			else:
				timespans.add(Stats.Timespan(span, unit))

		return timespans

	def lineify(self, data, full):
		"""
		Turns the data retrieved for a link into a list of lines.

		Each category has a header containing a string-representation of
		the category. For each timespan of a given category, a list is
		created if the category holds multiple data-points. Lines containing
		timespans and data-points are formatted into list-items using ecstasy.

		Arguments:
			data (dictionary): The data (statistics) to lineify. This data
							   should hold key/value pairs where the keys
							   are the sets/categories of data and the
							   values lists of dictionaries for each timespan.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').
		Returns:
			A list of lines, ready for output.
		"""
		lines = []
		for category, items in data.items():
			lines.append('{0}:'.format(category.title()))
			lines += self.listify(category, items, full)

		return lines

	def listify(self, category, data, full):
		"""
		Formats data with multiple data-points into a pretty list.

		Arguments:
			category (str): The category to which the data belongs
							(e.g. referrers).
			data (dict): The actual data, as a dictionary.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').
		Returns:
			A list of lines, ready for output.
		"""
		lines = []
		for result in data:
			timespan = result['timespan']
			items = result['data']
			header = self.get_header(timespan.span, timespan.unit)
			lines.append(header)
			if not items:
				lines[-1] += ' None'
			elif isinstance(items, list):
				for item in items:
					clicks = item.pop('clicks')
					key = list(item.values())[0]
					line = self.format(category, key, clicks, full)
					lines.append(line)
			else:
				# for clicks
				lines[-1] += ' {0}'.format(items)

		return lines

	def get_header(self, span, unit):
		"""
		Handles formatting of a timespan header.

		This method handles cases such as the 'since forever' timespan, as
		well as the issue with years (years have to be converted to months
		for bitly).

		Arguments:
			span (int): The span of the timespan (e.g. *4* months).
			unit (str): The unit of the timespan (e.g.. 4 *months*).

		Returns:
			A pretty list-item containing the timespan information.
		"""
		if span == -1:
			header = 'Since forever:'
		else:
			# Do this only for year because years have to be
			# converted to months before requesting stats from
			# the API, but it looks weird to the user if he
			# wanted years and got an equivalent number of months
			if unit == 'months' and span % 12 == 0:
				span //= 12
				unit = 'years' if span > 1 else 'year'
			span = '{0} '.format(span) if span > 1 else ''
			header = 'Last {0}{1}:'.format(span, unit)

		return self.list_item.format(header)

	@staticmethod
	def format(category, key, value, full):
		"""
		Formats a key/value pair for output.

		Handles special cases such as country-name expansion (the API returns
		them as ISO abbreviations, e.g. 'DE', but often the full name,
		e.g. 'Germany', is really wanted)

		Arguments:
			category (str): The category of the data.
			key (str): The key of the data-point.
			value (str): The value of the data-point.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').

		Returns:
			A pretty list-item.
		"""
		if category == 'countries':
			if key == 'None':
				key = 'Other'
			elif full:
				key = lnk.countries.names[key]
		elif key == 'direct':
			key = key.title()
		line = '   <-> {0}: {1}'.format(key, value)
		pretty = ecstasy.beautify(line, ecstasy.Color.Yellow)

		return pretty



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Contains the base-class for all goo.gl commands."""

import apiclient.discovery
import googleapiclient.errors
import httplib2
import oauth2client.file
import os

import lnk.config
import lnk.errors

from lnk.abstract import AbstractCommand

class Command(AbstractCommand):
	"""
	Base-class for all goo.gl commands.

	Configures the AbstractCommand base class for all commands in the
	entire application, which needs information about the service being
	used. Moreover fetches the oauth2 credentials for any HTTP request.

	Attributes:
		credentials (oauth2client.file.Storage): An object from Google's
												 API-library used to interact
												 with the oauth2 credentials
												 file.
	"""
	def __init__(self, which, credentials_path=None):
		"""
		Constructs a new Command.

		Arguments:
			which (str): The name of the command (e.g. link).
			credentials_path (str): Optionally, the full path to a credentials
									file. The one at lnk/config/credentials
									will be chosen by default.
		"""
		super(Command, self).__init__('googl', which)
		if credentials_path:
			self.credentials_path = credentials_path
		else:
			self.credentials_path = os.path.join(lnk.config.CONFIG_PATH,
												 '.credentials')
		self.credentials = oauth2client.file.Storage(self.credentials_path)

	def get_api(self):
		"""
		Returns an API-object used to perform any request.

		Returns:
			An API object from Google API-library, used to perform any
			HTTP request for the url-shortening API.
		"""
		http = self.authorize()
		api = apiclient.discovery.build('urlshortener', 'v1', http=http)

		return api.url()

	def get(self, url, projection=None, what=None):
		"""
		Base method to perform an HTTP request with the GET method.

		Arguments:
			url (str): The short goo.gl URL for which to perform a data-request.
			projection (str|None): Controls the amount of data returned for
								   the url (values should be either None,
								   such as for url-expansion, or 'FULL' to get
								   statistics or information about a link).
			what (str): A human-readable string representing what the request
						was for, such that if there is an error in the response,
						an errors.HTTPError is raised with the message
						'Could not <what>.' (NOT: 'even').

		Return:
			The requested data.
		"""
		api = self.get_api()
		request = api.get(shortUrl=url, projection=projection)
		response = self.execute(request, what)

		return response

	def authorize(self):
		"""
		Handles the authorization and re-authorization procedure.

		Normally, this method will return an HTTP object from Google's httplib2
		library, authorized to fetch user-specifc data by using the credentials
		that were configured for the user. There are two other cases: The oauth2
		access-token used to make oauth2 requests expires around-about every
		hour, so if that is the case this method will first refresh the token;
		if there are no credentials at all (i.e. the user did not run the 'key'
		command yet), this method throws an exception.

		Returns:
			If all goes well, an oauth2-authorized httplib2.HTTP object.

		Raises:
			errors.AuthorizationError: If there are no credentials for this user
									   yet (the initial authorization-procedure
									   was not yet performed).
		"""
		credentials = self.credentials.get()
		if not credentials:
			raise lnk.errors.AuthorizationError('goo.gl')
		http = httplib2.Http()
		if credentials.access_token_expired:
			credentials.refresh(http)
			self.credentials.put(credentials)
		credentials.authorize(http)

		return http

	@staticmethod
	def execute(request, what=None):
		"""
		Execute an HTTP request.

		The main point of this method is to catch HTTP errors raised by Google's
		API-library and re-raise them as lnk.errorrs.HTTPErrors.
		"""
		try:
			response = request.execute()
		except googleapiclient.errors.HttpError:
			raise lnk.errors.HTTPError('Could not {0}.'.format(what))

		return response



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""User-information from the bit.ly API."""

from __future__ import unicode_literals

import click
import time

try:
	from collections import OrderedDict
except ImportError:
	from ordereddict import OrderedDict

import lnk.abstract
import lnk.beauty
import lnk.bitly.history

from lnk.bitly.command import Command

def echo(*args):
	"""
	Executes a user command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 User instance's fetch() method.
	"""
	click.echo(User().fetch(*args))

class User(Command):
	"""
	Class to user-related information from the bit.ly API.

	User-related information includes data about since when the user
	is a member of bit.ly, the user's full name, the user's privacy
	settings, what share-accounts the user has associated his or her
	account with (e.g. Twitter or Facebook) and other things. Data
	may be put into a pretty string-box, or returned as a raw list
	for internal use.

	Attributes:
		raw (bool): Whether to prettify the output or
					return it raw, for internal use.
		history (bitly.history.History): An instance of bitly.history.History
										 to additionaly fetch link-history data
										 if so requested by the user.
		keys (dict): An additional mapping between the sets (data categories)
					 that the user can include from the command-line and proper
					 string-representations for output.
	"""
	def __init__(self, raw=False):
		super(User, self).__init__('user')
		self.raw = raw
		self.history = lnk.bitly.history.History(raw=True)
		# The point of this is that usually, the keys that the user can
		# supply from the command-line to address a certain data category
		# are fine for output, so there is just one mapping between those
		# categories and the actual names in the data returned by bitly.
		# This mapping is self.sets. Here, now, however, this additional
		# mapping between those category names and the outside-world
		# representation is needed.
		self.keys = {}
		for key, value in self.sets.items():
			if key == 'privacy':
				self.keys[value] = 'Link privacy'
			elif key == 'key':
				self.keys[value] = 'Api key'
			else:
				self.keys[value] = value.replace('_', ' ').title()

	def fetch(self, only, hide, _, add_history, hide_empty):
		"""
		Fetches user information.

		Arguments:
			only (tuple): A tuple of strings representing the sets to include
						  in the response ('only' these will be included).
			hide (tuple): A tuple of strings representing the sets to hide
						  from the response (either from all possible sets
						  if only is empty, or else from those selected).
			_ (bool): A dummy value that is actually the --all/--everything
					  flag for the command-line interface. It is needed from
					  the CLI if no specific data-sets should be included or
					  excluded (=> only/hide), but is of no use here because
					  if only and hide are empty no data-sets will be filtered
					  out anyway.
			add_history (bool): Whether to also show the user's link-history.
			hide_empty (bool): Whether or not to show things that are empty,
							   such as an empty list of share-accounts.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is True,
			else a boxified, pretty string.
		"""
		sets = lnk.abstract.filter_sets(self.sets, only, hide)
		data = self.request(sets.values())
		result = [self.lineify(data, hide_empty)]

		if add_history:
			result.append(self.get_history())
		elif self.raw:
			return result[0]

		return result if self.raw else lnk.beauty.boxify(result)

	def lineify(self, data, hide_empty):
		"""
		Formats the user-data retrieved into a list for output.

		Handles cases where the value for a certain key is a list, such as
		for the share-accounts, as well as if it is just a plain string. Also
		discards empty key/value pairs if hide_empty is true and the value
		is empty (empty iterable or string).

		Arguments:
			data (dict): The user-data retrieved.
			hide_empty (bool): Whether or not to show things that are empty,
							   such as an empty list of share-accounts.

		Returns:
			A list of pretty lines, ready for output.
		"""
		lines = []
		for key, value in data.items():
			if hide_empty and (not value and value != 0):
				continue
			if key == 'share_accounts':
				lines += self.format_accounts(value)
			elif isinstance(value, list):
				lines.append(self.format(key))
				lines += [self.list_item.format(i) for i in value]
			else:
				lines.append(self.format(key, value))

		return lines

	def format_accounts(self, value):
		"""
		Handles formatting of share-accounts.

		Share-accounts have a specific data-format that must be handled
		separately. For each share-account, a string-representation of the
		account-service is given (e.g. Twitter) and also the username for
		that service is displayed (@<username> for Twitter and the user's
		name for Facebook). Each share-account is finally shown as a string
		of the schema '<service>: <username>' and additionaly formatted
		into a pretty list-item.
		"""
		lines = ['Share Accounts:']
		if not value:
			lines[-1] += ' None'
		for account in value:
			if account['account_type'] == 'twitter':
				user = account['account_name']
				line = 'Twitter: @{0}'.format(user)
			elif account['account_type'] == 'facebook':
				user = account['account_name']
				line = 'Facebook: {0}'.format(user)
			else:
				line = account['account_type'].title()
			lines.append(self.list_item.format(line))

		return lines

	def format(self, key, value=''):
		"""
		Formats key/value pairs of user-data information.

		Handles special values and cases where the value is empty. Also
		parses and formats timestamps into proper (ctime) string-
		representations.

		Arguments:
			key (str): The key of the data-point.
			key (?): The value of the data-point.

		Returns:
			A string following the schema '<key>: <value>'.

		"""
		if not value and value != '':
			value = 'None'
		if key == 'member_since':
			value = time.ctime(value)

		return '{0}: {1}'.format(self.keys[key], value)

	def request(self, sets):
		"""
		Requests user-information from the bit.ly API.

		Arguments:
			sets (tuple): The data-categories to include.

		Returns:
			The data retrieved, ordered by the order() method.
		"""
		response = self.get(self.endpoints['user'])
		data = self.verify(response, 'retrieve user info')

		return self.order(data, sets)

	def order(self, data, sets):
		"""
		Partially-sorts user-data according to their priority.

		This method ensures that more important information such as the
		user's full name, his login or his membership-date is displayed
		at the very top of the data returned. This is necessary because
		all data is stored in a dictionary, which is unsorted by default.
		Key/value pairs other than the ones mentioned above are left in
		their original (unsorted) order.

		Arguments:
			data (dict): The raw, retrieved user-data.
			sets (tuple): The allowed categories of user-data.

		Returns:
			A partially-sorted collections.OrderedDict instance with the
			user's full-name in first position, then his or her username,
			then the date of his or her membership and then all other data
			in random order.
		"""
		ordered = OrderedDict()
		# Insert the ones we want on top first
		priority = ['full_name', 'login', 'member_since']
		for key in priority:
			if key in sets:
				ordered[key] = data.pop(key)

		for key, value in data.items():
			if key in sets:
				ordered[key] = value

		return ordered

	def get_history(self):
		"""
		Fetches a user's link-history.

		Uses a bitly.history.History instance.

		Returns:
			A list of links.
		"""
		links = self.history.fetch(None, None, True, None, False, False, False)

		return [self.list_item.format(link) for link in links]



# -*- coding: utf-8 -*-

"""
A mapping between short ISO abbreviations for countries and their full names.
"""

names = {

	'AF': 'Afghanistan',
	'AX': 'Åland Islands',
	'AL': 'Albania',
	'DZ': 'Algeria',
	'AS': 'American Samoa',
	'AD': 'Andorra',
	'AO': 'Angola',
	'AI': 'Anguilla',
	'AQ': 'Antarctica',
	'AG': 'Antigua and Barbuda',
	'AR': 'Argentina',
	'AM': 'Armenia',
	'AW': 'Aruba',
	'AU': 'Australia',
	'AT': 'Austria',
	'AZ': 'Azerbaijan',
	'BS': 'Bahamas',
	'BH': 'Bahrain',
	'BD': 'Bangladesh',
	'BB': 'Barbados',
	'BY': 'Belarus',
	'BE': 'Belgium',
	'BZ': 'Belize',
	'BJ': 'Benin',
	'BM': 'Bermuda',
	'BT': 'Bhutan',
	'BO': 'Bolivia, Plurinational State of',
	'BQ': 'Bonaire, Sint Eustatius and Saba',
	'BA': 'Bosnia and Herzegovina',
	'BW': 'Botswana',
	'BV': 'Bouvet Island',
	'BR': 'Brazil',
	'IO': 'British Indian Ocean Territory',
	'BN': 'Brunei Darussalam',
	'BG': 'Bulgaria',
	'BF': 'Burkina Faso',
	'BI': 'Burundi',
	'KH': 'Cambodia',
	'CM': 'Cameroon',
	'CA': 'Canada',
	'CV': 'Cape Verde',
	'KY': 'Cayman Islands',
	'CF': 'Central African Republic',
	'TD': 'Chad',
	'CL': 'Chile',
	'CN': 'China',
	'CX': 'Christmas Island',
	'CC': 'Cocos (Keeling) Islands',
	'CO': 'Colombia',
	'KM': 'Comoros',
	'CG': 'Congo',
	'CD': 'Congo, The Democratic Republic of the',
	'CK': 'Cook Islands',
	'CR': 'Costa Rica',
	'CI': "Côte d'Ivoire",
	'HR': 'Croatia',
	'CU': 'Cuba',
	'CW': 'Curaçao',
	'CY': 'Cyprus',
	'CZ': 'Czech Republic',
	'DK': 'Denmark',
	'DJ': 'Djibouti',
	'DM': 'Dominica',
	'DO': 'Dominican Republic',
	'EC': 'Ecuador',
	'EG': 'Egypt',
	'SV': 'El Salvador',
	'GQ': 'Equatorial Guinea',
	'ER': 'Eritrea',
	'EE': 'Estonia',
	'ET': 'Ethiopia',
	'FK': 'Falkland Islands (Malvinas)',
	'FO': 'Faroe Islands',
	'FJ': 'Fiji',
	'FI': 'Finland',
	'FR': 'France',
	'GF': 'French Guiana',
	'PF': 'French Polynesia',
	'TF': 'French Southern Territories',
	'GA': 'Gabon',
	'GM': 'Gambia',
	'GE': 'Georgia',
	'DE': 'Germany',
	'GH': 'Ghana',
	'GI': 'Gibraltar',
	'GR': 'Greece',
	'GL': 'Greenland',
	'GD': 'Grenada',
	'GP': 'Guadeloupe',
	'GU': 'Guam',
	'GT': 'Guatemala',
	'GG': 'Guernsey',
	'GN': 'Guinea',
	'GW': 'Guinea-Bissau',
	'GY': 'Guyana',
	'HT': 'Haiti',
	'HM': 'Heard Island and McDonald Islands',
	'VA': 'Holy See (Vatican City State)',
	'HN': 'Honduras',
	'HK': 'Hong Kong',
	'HU': 'Hungary',
	'IS': 'Iceland',
	'IN': 'India',
	'ID': 'Indonesia',
	'IR': 'Iran, Islamic Republic of',
	'IQ': 'Iraq',
	'IE': 'Ireland',
	'IM': 'Isle of Man',
	'IL': 'Israel',
	'IT': 'Italy',
	'JM': 'Jamaica',
	'JP': 'Japan',
	'JE': 'Jersey',
	'JO': 'Jordan',
	'KZ': 'Kazakhstan',
	'KE': 'Kenya',
	'KI': 'Kiribati',
	'KP': "Korea, Democratic People's Republic of",
	'KR': 'Korea, Republic of',
	'KW': 'Kuwait',
	'KG': 'Kyrgyzstan',
	'LA': "Lao People's Democratic Republic",
	'LV': 'Latvia',
	'LB': 'Lebanon',
	'LS': 'Lesotho',
	'LR': 'Liberia',
	'LY': 'Libya',
	'LI': 'Liechtenstein',
	'LT': 'Lithuania',
	'LU': 'Luxembourg',
	'MO': 'Macao',
	'MK': 'Macedonia, Republic of',
	'MG': 'Madagascar',
	'MW': 'Malawi',
	'MY': 'Malaysia',
	'MV': 'Maldives',
	'ML': 'Mali',
	'MT': 'Malta',
	'MH': 'Marshall Islands',
	'MQ': 'Martinique',
	'MR': 'Mauritania',
	'MU': 'Mauritius',
	'YT': 'Mayotte',
	'MX': 'Mexico',
	'FM': 'Micronesia, Federated States of',
	'MD': 'Moldova, Republic of',
	'MC': 'Monaco',
	'MN': 'Mongolia',
	'ME': 'Montenegro',
	'MS': 'Montserrat',
	'MA': 'Morocco',
	'MZ': 'Mozambique',
	'MM': 'Myanmar',
	'NA': 'Namibia',
	'NR': 'Nauru',
	'NP': 'Nepal',
	'NL': 'Netherlands',
	'NC': 'New Caledonia',
	'NZ': 'New Zealand',
	'NI': 'Nicaragua',
	'NE': 'Niger',
	'NG': 'Nigeria',
	'NU': 'Niue',
	'NF': 'Norfolk Island',
	'MP': 'Northern Mariana Islands',
	'NO': 'Norway',
	'OM': 'Oman',
	'PK': 'Pakistan',
	'PW': 'Palau',
	'PS': 'Palestine, State of',
	'PA': 'Panama',
	'PG': 'Papua New Guinea',
	'PY': 'Paraguay',
	'PE': 'Peru',
	'PH': 'Philippines',
	'PN': 'Pitcairn',
	'PL': 'Poland',
	'PT': 'Portugal',
	'PR': 'Puerto Rico',
	'QA': 'Qatar',
	'RE': 'Réunion',
	'RO': 'Romania',
	'RU': 'Russian Federation',
	'RW': 'Rwanda',
	'BL': 'Saint Barthélemy',
	'SH': 'Saint Helena, Ascension and Tristan da Cunha',
	'KN': 'Saint Kitts and Nevis',
	'LC': 'Saint Lucia',
	'MF': 'Saint Martin (French part)',
	'PM': 'Saint Pierre and Miquelon',
	'VC': 'Saint Vincent and the Grenadines',
	'WS': 'Samoa',
	'SM': 'San Marino',
	'ST': 'Sao Tome and Principe',
	'SA': 'Saudi Arabia',
	'SN': 'Senegal',
	'RS': 'Serbia',
	'SC': 'Seychelles',
	'SL': 'Sierra Leone',
	'SG': 'Singapore',
	'SX': 'Sint Maarten (Dutch part)',
	'SK': 'Slovakia',
	'SI': 'Slovenia',
	'SB': 'Solomon Islands',
	'SO': 'Somalia',
	'ZA': 'South Africa',
	'GS': 'South Georgia and the South Sandwich Islands',
	'ES': 'Spain',
	'LK': 'Sri Lanka',
	'SD': 'Sudan',
	'SR': 'Suriname',
	'SS': 'South Sudan',
	'SJ': 'Svalbard and Jan Mayen',
	'SZ': 'Swaziland',
	'SE': 'Sweden',
	'CH': 'Switzerland',
	'SY': 'Syrian Arab Republic',
	'TW': 'Taiwan, Province of China',
	'TJ': 'Tajikistan',
	'TZ': 'Tanzania, United Republic of',
	'TH': 'Thailand',
	'TL': 'Timor-Leste',
	'TG': 'Togo',
	'TK': 'Tokelau',
	'TO': 'Tonga',
	'TT': 'Trinidad and Tobago',
	'TN': 'Tunisia',
	'TR': 'Turkey',
	'TM': 'Turkmenistan',
	'TC': 'Turks and Caicos Islands',
	'TV': 'Tuvalu',
	'UG': 'Uganda',
	'UA': 'Ukraine',
	'AE': 'United Arab Emirates',
	'GB': 'United Kingdom',
	'US': 'United States',
	'UM': 'United States Minor Outlying Islands',
	'UY': 'Uruguay',
	'UZ': 'Uzbekistan',
	'VU': 'Vanuatu',
	'VE': 'Venezuela, Bolivarian Republic of',
	'VN': 'Viet Nam',
	'VG': 'Virgin Islands, British',
	'VI': 'Virgin Islands, U.S.',
	'WF': 'Wallis and Futuna',
	'EH': 'Western Sahara',
	'YE': 'Yemen',
	'ZM': 'Zambia',
	'ZW': 'Zimbabw',
}



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

import os
import json
import pytest

from collections import namedtuple

import tests.paths

import lnk.config
import lnk.errors

HERE = os.path.dirname(os.path.abspath(__file__))

@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'which',
		'file',
		'manager',
		'config'
		])

	which = 'test_manager'
	filename = '{0}.json'.format(which)
	path = os.path.join(tests.paths.CONFIG_PATH, filename)

	with open(os.path.join(HERE, filename)) as source:
		config = json.load(source)
		with open(path, 'w') as test:
			json.dump(config, test)

	def finalize():
		os.remove(path)

	request.addfinalizer(finalize)

	manager = lnk.config.Manager()

	return Fixture(which, path, manager, config)

@pytest.fixture(scope='module')
def changed():
	Fixture = namedtuple('Fixture', ['contents', 'config'])
	with open(os.path.join(HERE, 'changed_manager.json')) as source:
		contents = source.read()
		config = json.loads(contents)

	return Fixture(contents, config)



def test_opens_correctly(fixture):
	fixture.manager.open(fixture.which)

	assert fixture.manager.file == fixture.file
	assert fixture.manager.config == fixture.config


def test_writes_correctly(fixture, changed):
	fixture.manager['fucks'] = -1
	fixture.manager['animal'] = 'unicorn'

	fixture.manager.write()

	with open(fixture.file) as test:
		assert json.load(test) == changed.config


def test_throws_for_invalid_key(fixture):
	with pytest.raises(lnk.errors.InvalidKeyError):
		fixture.manager['random'] = None


def test_context_syntax_closes_well(fixture):
	with lnk.config.Manager(fixture.which) as manager:
		assert manager['fucks'] == -1
		assert manager['animal'] == 'unicorn'
		manager['fucks'] = 0
		manager['animal'] = 'cat'
		assert manager.config == fixture.config
		manager.write()

	with open(fixture.file) as test:
		assert json.load(test) == fixture.config


def test_context_syntax_writes_well(fixture, changed):
	with lnk.config.Manager(fixture.which, write=True) as manager:
		assert manager['fucks'] == 0
		assert manager['animal'] == 'cat'
		manager['fucks'] = -1
		manager['animal'] = 'unicorn'
		assert manager.config == changed.config

	with open(fixture.file) as test:
		assert json.load(test) == changed.config


def test_properties_are_accessible(fixture, changed):
	assert all(i in fixture.manager.keys for i in changed.config.keys())
	assert all(i in fixture.manager.values for i in changed.config.values())
	assert all(i in fixture.manager.items for i in changed.config.items())


def test_get_works(changed):
	result = lnk.config.get('test_manager', 'animal')
	expected = changed.config['animal']

	assert result == expected


def test_closes_correctly(fixture):
	fixture.manager.close()

	assert fixture.manager.file is None
	assert fixture.manager.config is None


def test_throws_for_write_when_no_file_open(fixture):
	with pytest.raises(lnk.errors.InternalError):
		fixture.manager.write()


def test_throws_for_close_when_no_file_open(fixture):
	with pytest.raises(lnk.errors.InternalError):
		fixture.manager.close()



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Statistics and metrics retrieval for goo.gl."""

from __future__ import unicode_literals

import click
import ecstasy

import lnk.abstract
import lnk.beauty
import lnk.countries

from lnk.googl.command import Command

def echo(*args):
	"""
	Executes a stats command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 Stats instance's fetch() method.
	"""
	click.echo(Stats().fetch(*args))

class Stats(Command):
	"""
	Class to retrieve statistics and info for one or more goo.gl links.

	The statistics for a link include its referrers (i.e. from where the link
	was opened), the countries from where the link was opened, the browsers in
	which the link was opened, the platforms (operating systems) from which the
	link was opened and of course the number of clicks. These statistics can be
	retrieved 'since-forever', but also for specific (possibly open-ended)
	time-ranges, such as for 'the last 5 months' or 'between 4 days and 2 minute
	ago'. Additionally, these statistics can be paired with information about
	each link, retrieved from the 'info' command, thereby making the stats
	command the ultimate destination for link statistics *and* information.
	Output may, as always, be in raw format for internal use or in a pretty box.
	Multiple URLs are fully supported.

	Attributes:
		raw (bool): Whether to return the output in raw format for internal use,
					or in a pretty string-representation for outside-display.
	"""
	def __init__(self, raw=False):
		super(Stats, self).__init__('stats')
		self.raw = raw

	def fetch(self, only, hide, times, forever, limit, add_info, full, urls):
		"""
		Fetches statistics for one or more URLs.

		Arguments:
			only (tuple): A tuple of strings representing the sets to include
						  in the response ('only' these will be included).
			hide (tuple): A tuple of strings representing the sets to hide
						  from the response (either from all possible sets
						  if only is empty, or else from those selected).
			times (tuple): A tuple of tuples of the schema (<span>, <unit>),
						   representing the timespans for which to fetch
						   statistics.
			forever (bool): Whether to include the statistics 'since forever'.
			limit (int): A limit on the number of items fetched per timespan.
			add_info (bool): Whether or not to add additional information for
							 each link.
			full (bool): Whether to show full country names, or short ISO
						 abbreviations.
			urls (tuple): A tuple of urls to fetch statistics for.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is True,
			else a boxified, pretty string.
		"""
		sets = lnk.abstract.filter_sets(self.sets, only, hide)
		timespans = self.get_timespans(times, forever)

		results = []
		threads = []
		for url in urls:
			self.queue.put(url)
			args = (results, sets, timespans, add_info, full, limit)
			thread = self.new_thread(self.get_stats, *args)
			threads.append(thread)
		self.join(threads)

		return results if self.raw else lnk.beauty.boxify(results)

	def get_stats(self, results, sets, timespans, add_info, full, limit):
		"""
		Retrieves the statistics for a single url.

		The statistics returned are for all timespans supplied, filtered
		according to the sets of statistics wanted.

		Arguments:
			results (list): The list of results to which to append the
							retreived data.
			sets (tuple): The sets of statistics wanted in the response
						 (others are discarded).
			timespans (tuple): A tuple of tuples of the schema (<span>, <unit>),
							   representing the timespans for which to fetch
							   statistics.
			add_info (bool): Whether to add information about each url
							 alongside statistics.
			full (bool): Whether to show full country names or short ISO
						 abbreviations.
			limit (int): A limit to the number of data-points selected for each
						 timespan.
		"""
		data = self.request(add_info)
		lines = self.lineify(data, sets, timespans, full, limit)
		with self.lock:
			results.append(lines)

	def request(self, add_info):
		"""
		Requests statistics for a given configuration.

		The url for which to retrieve statistics is fetched from the queue
		attribute (thread-safe).

		Arguments:
			add_info (bool): Whether to add information about the URL as well.

		Returns:
			The dictionary containing the retrieved data. The format of this
			dictionary is that it contains a 'URL' key, as well as all key/value
			pairs connected to information. It also contains an 'analytics' key,
			whose value are the actual statistics for the URL. These statistics
			are a dictionary where the keys are the timespans and the values
			the data for each timespan. This data is then furthermore a
			dictionary which maps the category/set names to list of data-points.
			Each data-point's key is contained in the 'id' key (e.g. the name
			of a country) and the value (always the number of clicks) is
			found with the 'count' key.

		"""
		url = self.queue.get()
		what = "get information for '{0}'".format(url)
		data = self.get(url, 'FULL', what)

		data['URL'] = url
		del data['kind']
		del data['id']
		if not add_info:
			for i in ['created', 'longUrl', 'status']:
				del data[i]

		return data

	def lineify(self, data, sets, timespans, full, limit):
		"""
		Turns the data retrieved for a link into a list of lines.

		Information about the URL is formatted and turned into header-lines
		here, the statistics are entirely parsed in the listify() method.

		Arguments:
			data (dictionary): The data (statistics) to lineify.
			sets (tuple): The names of the sets to include in the output.
			timespans (tuple): The timespans for which to include data.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').
			limit (int): A limit on the number of data-points retrieved per
						 timespan.
		Returns:
			A list of lines, ready for output.
		"""
		stats = data.pop('analytics')
		statistics = self.listify(stats, sets, timespans, full, limit)
		header = [self.format(key, value) for key, value in data.items()]

		return header + statistics

	def listify(self, data, sets, timespans, full, limit):
		"""
		Formats data with multiple data-points into a pretty list.

		Each category is formatted into a header. For each timespan of a given
		category, a list is created if the category holds multiple data-points.
		Lines containing timespans and data-points are formatted into list-items
		using ecstasy.

		Arguments:
			data (dictionary): The data (statistics) to lineify.
			sets (tuple): The names of the sets to include in the output.
			timespans (tuple): The timespans for which to include data.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').
			limit (int): A limit on the number of data-points retrieved per
						 timespan.
		Returns:
			A list of lines, ready for output.
		"""
		lines = []
		for display, real in sets.items():
			lines.append('{0}:'.format(display.title()))
			for timespan, categories in data.items():
				# Ignore unwanted timespans
				if timespan not in timespans:
					continue
				lines.append(self.get_header(timespan))
				# The goo.gl API does not include categories with zero
				# clicks thus we first have to determine whether the
				# category is present at all
				if not categories.get(real):
					lines[-1] += ' None'
				elif display == 'clicks':
					lines[-1] += ' {0}'.format(categories[real])
				else:
					lines += self.sub_listify(display,
											  categories[real],
											  limit,
											  full)
		return lines

	def get_header(self, timespan):
		"""
		Formats timespan headers.

		Among other things, this method takes care of the 'forever' timespan,
		whose name to the goo.gl API is 'allTime'. Before returning a header,
		it is formatted using ecstasy into a proper, pretty list-item.
		"""
		if timespan == 'allTime':
			header = 'Since forever:'
		else:
			if timespan == 'twoHours':
				timespan = 'two hours'
			header = 'Last {0}:'.format(timespan)

		return self.list_item.format(header)

	def sub_listify(self, category, points, limit, full):
		"""
		Handles transforming data for a category into a level-2 list.

		While listify handles all data and takes care of formatting timespans,
		this method handles only data for a single category. Because this data
		is on the second level of each list (0th level is the category, 1st
		is the timespan and 2nd are the data-points), it is formatted
		differently than the data on level-1 (the level-1 bullet is a '+',
		the level-2 bullet a '-'). This method also handles special cases
		for key-names and also country-name expansion.

		Arguments:
			category (str): The name of the category of this data.
			points (list): The data points for this category (the list
						   contains dictionaries with the keys 'id' and
						   'count').
			limit (int): A limit on the number of data-points retrieved per
						 timespan.
			full (bool): Whether to show full country names or short
						 ISO abbreviations (e.g. 'Germany' or 'DE').
		"""
		lines = []
		for n, point in enumerate(points):
			if n == limit:
				break
			subject = point['id']
			if subject == 'unknown':
				subject = subject.title()
			if category == 'countries' and full:
				subject = lnk.countries.names[subject]
			clicks = point['count']
			line = ecstasy.beautify('   <-> {0}: {1}', ecstasy.Color.Yellow)
			lines.append(line.format(subject, clicks))

		return lines

	def get_timespans(self, times, forever):
		"""
		Parses timespans.

		The timespans passed to the command-line interface and ultimately
		to the fetch() method are of the schema '(<span>, <unit>)'. These
		are parsed such that special cases are handled, e.g. if the 'forever'
		flag is set, the timespan 'allTime' is added for the goo.gl API.
		If no timespans were passed to the method, the default timespan is
		retrieved from the command's configuration settings.

		Arguments:
			times (tuple): The timespans of the schema '(<span>, <unit>)'.
			forever (bool): Whether to include the 'since forever' timespan.

		Returns:
			A set of timespans (string-representations thereof).
		"""
		timespans = set(times)
		if not timespans and not forever:
			default = self.settings['timespan']
			if default == 'forever':
				timespans.add('allTime')
			elif default == 'two-hours':
				timespans.add('twoHours')
			else:
				timespans.add(default)
		if forever:
			timespans.add('allTime')
		if 'two-hours' in timespans:
			timespans.remove('two-hours')
			timespans.add('twoHours')

		return timespans

	@staticmethod
	def format(key, value):
		"""
		Formats a key/value pair for output.

		Handles special cases for key-names and formats the key-value pair.

		Arguments:
			key (str): The key of the data-point.
			value (str): The value of the data-point.

		Returns:
			A pretty key-value string-representation.
		"""
		if key == 'shortUrlClicks':
			key = 'clicks'
		elif key == 'longUrl':
			key = 'expanded'

		return '{0}: {1}'.format(key.title(), value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import ecstasy
import pytest

import tests.paths

import lnk.beauty

@pytest.fixture(scope='module')

def fixture():
	leg = ecstasy.beautify('<leg>', ecstasy.Style.Blink | ecstasy.Color.Blue)
	return [[leg, 'grapefruit'], ['hippopotamus']]


def test_max_width_60_if_no_terminal():
	assert lnk.beauty.MAX_WIDTH == 60

	# What we'll use for testing
	lnk.beauty.MAX_WIDTH = 10


def test_escape_escapes_well(fixture):
	result = lnk.beauty.escape(fixture[0][0])

	assert isinstance(result, tuple)
	assert result.raw == fixture[0][0]
	assert result.escaped == 'leg'


def test_escape_leaves_strings_w_o_escape_codes_alone(fixture):
	result = lnk.beauty.escape(fixture[0][1])

	assert isinstance(result, tuple)
	assert result.raw == fixture[0][1]
	assert result.escaped == fixture[0][1]


def test_get_escaped_escapes_results_well(fixture):
	result, _ = lnk.beauty.get_escaped(fixture)
	expected = [[lnk.beauty.escape(j) for j in i] for i in fixture]

	assert result == expected


def test_get_escaped_picks_correct_width(fixture):
	_, width = lnk.beauty.get_escaped([fixture[0]])

	assert width == len('grapefruit')


def test_get_escaped_truncates_to_max_width(fixture):
	_, width = lnk.beauty.get_escaped(fixture)

	assert width == lnk.beauty.MAX_WIDTH


def test_ljust_works_well(fixture):
	line = lnk.beauty.escape(fixture[0][0])
	adjusted = lnk.beauty.ljust(line, 10)
	expected = '{0}{1}'.format(fixture[0][0], ' ' * 7)

	assert adjusted == expected


def test_wrap_works_well(fixture):
	line = lnk.beauty.escape(fixture[0][0])
	wrapped = [i.escaped for i in lnk.beauty.wrap(line, 2)]

	assert len(wrapped) == 2
	assert not any(len(i) > 10 for i in wrapped)
	assert wrapped == ['le', ' g']


def test_boxify_works_well(fixture):
	lines = [lnk.beauty.escape(j) for i in fixture for j in i]
	border = '─' * (lnk.beauty.MAX_WIDTH + 2)
	wrapped = lnk.beauty.wrap(lines[2], lnk.beauty.MAX_WIDTH)
	width = lnk.beauty.MAX_WIDTH
	expected = [
		'┌{0}┐'.format(border),
		'│ {0} │'.format(lnk.beauty.ljust(lines[0], width)),
		'│ {0} │'.format(lnk.beauty.ljust(lines[1], width)),
		'├{0}┤'.format(border),
		'│ {0} │'.format(lnk.beauty.ljust(wrapped[0], width)),
		'│ {0} │'.format(lnk.beauty.ljust(wrapped[1], width)),
		'└{0}┘'.format(border)
	]
	print(fixture)
	box = lnk.beauty.boxify(fixture)
	print(box)
	assert box == '\n'.join(expected)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Link shortening and expansion for the goo.gl client."""

import apiclient.discovery
import click
import ecstasy
import pyperclip
import re

import lnk.beauty
import lnk.errors

from lnk.googl.command import Command

def echo(*args):
	"""
	Executes a Link command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 Link instance's fetch() method.
	"""
	click.echo(Link().fetch(*args))

class Link(Command):
	"""
	Class to shorten or expand a url using goo.gl.

	This class can shorten a long url to a shortened goo.gl url,
	or expand a shortened goo.gl url to its original long url.

	Attributes:
		raw (bool): Whether to prettify the output or
					return it raw, for internal use.
		already_copied (bool): Flag set the first time a url is copied
							   to the clipboard, such that only the first
							   is copied.
		http (regex): A compiled regular-expression object matching a
					  HTTP(S) protocol, for URL-checking.
	"""

	def __init__(self, raw=False):
		super(Link, self).__init__('link')
		self.raw = raw
		self.already_copied = False
		self.http = re.compile(r'https?://')

	def fetch(self, copy, quiet, expand, shorten, pretty):
		"""
		Fetches a short or expanded link.

		Arguments:
			copy (bool): Whether or not to copy the first shortened or
						 expanded link to the clipboard.
			quiet (bool): Whether or not to swallow warnings about urls
						  having to be modified because they do not have
						  a protocol specified (i.e. no 'http://').
			expand (tuple): A tuple of short urls to expand.
			shorten (tuple): A tuple of long urls to shorten.
			pretty (bool): Whether or not to make the output pretty (in a box).

		Returns:
			A raw list of lines for other internal use if self.raw is True,
			else a pretty list in a box if pretty is True, else the same
			list as in the first case, but joined to a string for output.
		"""
		self.already_copied = False
		result = self.shorten_urls(copy, quiet, shorten)
		result += self.expand_urls(copy, expand)

		if self.raw:
			return result
		return lnk.beauty.boxify([result]) if pretty else '\n'.join(result)

	def shorten_urls(self, copy, quiet, urls):
		"""
		Shortens a sequence of long urls.

		Arguments:
			copy (bool): Whether or not to copy the first url to the clipboard.
			quiet (bool): Whether or not to swallow warnings about urls
						  having to be modified because they do not have
						  a protocol specified (i.e. no 'http://').
			urls (tuple): The tuple of long urls to shorten.

		Returns:
			A list of lines for output.
		"""
		lines = []
		threads = []
		for url in urls:
			if not self.http.match(url):
				url = 'http://{0}'.format(url)
				if not quiet:
					lnk.errors.warn("Prepending 'http://' to '{0}'".format(url))
			self.queue.put(url)
			threads.append(self.new_thread(self.shorten, lines, copy))
		self.join(threads)

		return lines

	def expand_urls(self, copy, urls):
		"""
		Expands a sequence of short urls.

		Arguments:
			copy (bool): Whether or not to copy the first url to the clipboard.
			urls (tuple): The tuple of short urls to expand.

		Returns:
			A list of lines for output.
		"""
		lines = []
		threads = []
		for url in urls:
			self.queue.put(url)
			threads.append(self.new_thread(self.expand, lines, copy))
		self.join(threads)

		return lines

	def shorten(self, lines, copy):
		"""
		Requests a shortend link and appends it to a given list.

		A shortened link is retrieved via get_short(), then (possibly)
		copied to the clipboard via copy(), then appended to the given
		list in a thread-safe (locked) way.

		Arguments:
			lines (list): The list of lines to append the url to.
			copy (bool): Whether or not to copy the link to the clipboard
		"""
		url = self.queue.get()
		short = self.get_short(url)
		formatted = self.copy(copy, short)
		with self.lock:
			lines.append('{0} => {1}'.format(url, formatted))

	def expand(self, lines, copy):
		"""
		Requests an expanded link and appends it to a given list.

		An expanded link is retrieved via get_long(), then (possibly)
		copied to the clipboard via copy(), then appended to the given
		list in a thread-safe (locked) way.

		Arguments:
			lines (list): The list of lines to append the url to.
			copy (bool): Whether or not to copy the link to the clipboard
						 (if no other link has already been copied).
		"""
		url = self.queue.get()
		expanded = self.get_long(url)
		formatted = self.copy(copy, expanded)
		with self.lock:
			lines.append('{0} => {1}'.format(url, formatted))

	def get_short(self, url):
		"""
		Requests and returns a short url for a long one.

		Arguments:
			url (str): The long url to shorten.

		Returns:
			The shortened link.
		"""
		api = self.get_api()
		request = api.insert(body=dict(longUrl=url))
		what = "shorten url '{0}'".format(url)
		response = self.execute(request, what)

		return response['id']

	def get_long(self, url):
		"""
		Requests and returns an expanded url for a short one.

		Arguments:
			url (str): The short goo.gl link (bitlink) to expand.

		Returns:
			The expanded link.
		"""
		response = self.get(url, what="expand url '{0}'".format(url))

		if response['status'] in ['MALWARE', 'PHISHING']:
			what = "Careful! goo.gl believes the url '{0}' is {1}!"
			lnk.errors.warn(what.format(response['longUrl'],
										response['status'].lower()))
		elif response['status'] == 'REMOVED':
			return 'REMOVED'

		return response['longUrl']

	def copy(self, copy, url):
		"""
		Copies a url to the clipboard if possible.

		"If possible" means the url is only copied if no other url
		has so-far been copied to the clipboard during the same
		fetch() call, and also only if the 'copy' is set. If the url
		is copied, it is formatted using ecstasy to make it bold.

		Arguments:
			copy (bool): Whether or not to allow copying
						 to the clipboard at all.
			url (str): The url to (possibly) copy.

		Returns:
			The original url if the 'copy' flag is unset or another link
			has already been copied to the clipboard during the same call
			to fetch(), else the url, formatted with ecstasy to appear bold
			in the terminal.
		"""
		if copy and not self.already_copied:
			self.already_copied = True
			pyperclip.copy(url)
			url = ecstasy.beautify('<{0}>'.format(url), ecstasy.Style.Bold)

		return url



Python 3.5.0 (default, Sep 14 2015, 02:37:27) 
[GCC 4.2.1 Compatible Apple LLVM 6.1.0 (clang-602.0.53)] on darwin
Type "help", "copyright", "credits" or "license" for more information.
>>> a = [{'a': 1}, {'b': 2}]
>>> sorted(a) 
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
TypeError: unorderable types: dict() < dict()
>>> b = [{'b': 2}, {'a': 1}]


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import click
import ecstasy
import pytest
import requests

import tests.paths
import lnk.errors


def test_verbosity_system_works_without_additional():
	what = 'something happened'
	error = lnk.errors.Error(what)
	typ = ecstasy.beautify('<Type>: Error', ecstasy.Color.Red)

	assert error.what == what
	assert error.levels[2] == typ


def test_get_levels_works():
	error = lnk.errors.Error('something happened')

	assert len(error.levels) == 4
	assert any(error.levels)
	assert not error.levels[1]
	assert not error.levels[3]
	assert 'Error' in error.levels[0]
	assert 'Type' in error.levels[2]


def test_verbosity_system_works_with_additional():
	foo = lnk.errors.Message(what='foo?', level=1)
	bar = lnk.errors.Message(what='bar!', level=3)
	error = lnk.errors.Error('something happened', Foo=foo, Bar=bar)

	assert len(error.levels) == 4
	assert all(error.levels)
	assert 'Foo' in error.levels[1]
	assert 'foo?' in error.levels[1]
	assert 'Bar' in error.levels[3]
	assert 'bar!' in error.levels[3]


def test_catch_catches_lnk_error(capsys):
	def throws():
		raise lnk.errors.Error('oops')
	lnk.errors.catch(throws)
	captured = capsys.readouterr()

	assert captured
	assert 'Error' in captured[0]
	assert 'oops' in captured[0]


def test_catch_shows_only_wanted_levels_for_verbosity_0(capsys):
	catch = lnk.errors.Catch()
	def throws():
		raise lnk.errors.Error('oops')
	catch.catch(throws)
	captured = capsys.readouterr()

	assert captured

	levels = [i for i in captured[0].split('\n') if i]

	assert len(levels) == 1
	assert 'Error' in levels[0]
	assert 'oops' in levels[0]


def test_catch_shows_all_levels_for_verbosity_4(capsys):
	catch = lnk.errors.Catch(3)
	def throws():
		foo = lnk.errors.Message(what='foo?', level=1)
		bar = lnk.errors.Message(what='bar!', level=3)
		raise lnk.errors.InternalError('oops', Foo=foo, Bar=bar)
	catch.catch(throws)
	captured = capsys.readouterr()

	assert captured

	levels = [i for i in captured[0].split('\n') if i]

	assert len(levels) == 4
	assert 'Error' in levels[0]
	assert 'oops' in levels[0]
	assert 'Foo' in levels[1]
	assert 'foo?' in levels[1]
	assert 'Type' in levels[2]
	assert 'InternalError' in levels[2]
	assert 'Bar' in levels[3]
	assert 'bar!' in levels[3]


def test_catch_catches_click_exception(capsys):
	catch = lnk.errors.Catch(2)
	def throws():
		raise click.ClickException('')
	catch.catch(throws)
	captured = capsys.readouterr()

	assert captured

	levels = [i for i in captured[0].split('\n') if i]

	assert 'Error' in levels[0]
	assert 'Type' in levels[1]
	assert 'UsageError' in levels[1]


def test_catch_catches_requests_exception(capsys):
	catch = lnk.errors.Catch(2)
	def throws():
		raise requests.exceptions.ConnectionError
	catch.catch(throws)
	captured = capsys.readouterr()

	assert captured

	levels = [i for i in captured[0].split('\n') if i]

	assert 'Error' in levels[0]
	assert 'Type' in levels[2]
	assert 'ConnectionError' in levels[2]


def test_catch_bubbles_up_other_exceptions():
	catch = lnk.errors.Catch()
	def throws():
		raise RuntimeError

	with pytest.raises(RuntimeError):
		catch.catch(throws)


def test_warn_works(capsys):
	lnk.errors.warn('Sauron is coming')
	captured = capsys.readouterr()

	assert captured
	assert 'Warning' in captured[0]
	assert '\a' in captured[0]
	assert 'Sauron is coming' in captured[0]



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Link-history-retrieval for the goo.gl client."""

from __future__ import unicode_literals

import apiclient.discovery
import click
import ecstasy
import warnings

from collections import namedtuple
from datetime import datetime, timedelta

import lnk.beauty
import lnk.config
import lnk.errors

from lnk.googl.command import Command

warnings.filterwarnings('ignore', module=r'ecstasy\.parser')

def echo(*args):
	"""
	Executes a history command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to a
						 History instance's fetch() method.
	"""
	click.echo(History().fetch(*args))

class History(Command):

	"""
	Class to retrieve goo.gl link history for a user.

	The sole purpose of this class is to fetch and return a list of
	urls the user has shortened using goo.gl in the past. This list
	may, of course, be properly prettified if necessary. A nice feature
	is that if the history is fetched with the 'plain' flag set to true
	its output can be piped into other lnk commands, such as 'stat' or 'info'.

	Attributes:
		raw (bool): Whether to prettify the output or
					return it raw, for internal use.
		delta (dict): A mapping between string representations of time
						and equivalent datetime.timedelta objects.
		Url (namedtuple): Class-attribute namedtuple to represent a link
						  with a short, long/expanded url and a creation
						  date.

	Note:
		The datetime.timedelta objects store time-offsets as floating-point
		values (e.g. 1 year = 365.242199 days) to account for time-weirdness
		such as shift-years, but they will nevertheless not be absolutely
		precise, but good enough.
	"""

	Url = namedtuple('Url', ['short', 'long', 'created'])

	def __init__(self, raw=False):
		super(History, self).__init__('history')
		self.raw = raw
		self.delta = {
			'minute': timedelta(minutes=1), 
			'hour': timedelta(hours=1), 
			'day': timedelta(days=1),
			'week': timedelta(weeks=1), 
			'month': timedelta(days=30.4368),
			'year': timedelta(days=365.242199)
		}

	def fetch(self, last, ranges, forever, limit, expanded, both, pretty):
		"""
		Fetches the link history and returns a string for output.

		Arguments:
			last (tuple): A tuple of time_points to retrieve history for, without
						  an upper bound (i.e. all links after a given time_point,
						  up to now).
			ranges (tuple): A tuple of time_point-ranges to retrieve history for
							(with a lowe and upper bound).
			forever (bool): Whether or not to include data for all
							time_points (since forever).
			limit (int): A limit on the number of datapoints fetched per time_point.
			expanded (bool): Whether or not to show expanded rather
							 than short links.
			both (bool): Whether or not to show both short and expanded links
						 (takes precedence over 'expanded').
			pretty (bool): Whether or not to prettify the output.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is true,
			else a boxified, pretty string if the pretty flag is set,
			else the same plain list as for the first case, but joined to
			a string.
		"""
		data = self.request()
		urls = self.process(data)

		result = []
		if forever:
			result += self.forever(urls, limit, expanded, both, pretty)
		if ranges:
			result += self.ranges(urls, set(ranges), limit, expanded, both, pretty)
		if last:
			result += self.last(urls, set(last), limit, expanded, both, pretty)

		# Remove last empty line
		if pretty:
			del result[-1]

		if self.raw:
			return result
		return lnk.beauty.boxify([result]) if pretty else '\n'.join(result)

	def forever(self, urls, limit, expanded, both, pretty):
		"""
		Formats history for all possible time-ranges.

		Arguments:
			urls (list|tuple): The original/full list of urls.
			limit (int): A limit to the number of links to pick.
			expanded (bool): Whether to show expanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header ('Since forever:') plus
			an empty line for padding.
		"""
		lines = []
		for n, url in enumerate(urls):
			if n == limit:
				break
			line = self.lineify(url, expanded, both, pretty)
			lines.append(line)

		return ['Since forever:'] + lines + [''] if pretty else lines

	def ranges(self, urls, ranges, limit, expanded, both, pretty):
		"""
		Formats and filters history for certain time ranges.

		Arguments:
			urls (list|tuple): The original/full list of urls.
			ranges (tuple): The time-ranges of schema (span1, unit1, span2, unit2).
			limit (int): A limit to the number of links to pick for each time-range.
			expanded (bool): Whether to show expanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header (e.g. 'Between 7 months and
			5 days ago:'), plus an empty line for padding.
		"""
		lines = []
		for time_range in ranges:
			begin, end = self.get_boundaries(time_range)
			filtered = self.filter(urls, begin, end)
			if pretty:
				header = self.ranges_header(time_range, filtered)
				lines.append(header)
			lines += self.listify(filtered, limit, expanded, both, pretty)
			if pretty:
				lines.append('')

		return lines

	def last(self, urls, last, limit, expanded, both, pretty):
		"""
		Fetches and formats history after a given time-point.

		Arguments:
			urls (list|tuple): The original/full list of urls.
			last (tuple): The open-ended time-ranges of schema (span, unit).
			limit (int): A limit to the number of links to pick for each
						 time-range.
			expanded (bool): Whether to show exapanded or short links.
			both (bool): Whether to show both expanded and short links.
			pretty (bool): Whether to prettify the output.

		Returns:
			A plain list of lines ready if pretty is false, else the
			same list of lines with a header (e.g. 'Last 4 weeks:'),
			plus an empty line for padding.
		"""
		lines = []
		for time_point in last:
			begin = self.get_date(time_point)
			filtered = self.filter(urls, begin, datetime.now())
			if pretty:
				header = self.last_header(time_point, filtered)
				lines.append(header)
			lines += self.listify(filtered, limit, expanded, both, pretty)
			if pretty:
				lines.append('')

		return lines

	def ranges_header(self, time_range, urls):
		"""
		Returns a header for a time-range.

		Arguments:
			time_range (tuple): The time-range following the
								schema (span1, unit1, span2, unit2).
			urls (list|tuple): The urls, to check if empty.

		Returns:
			A header of the schema 'Between <span1> [<unit1>] and <span2>
			<unit2>', where <unit1> is only included if it not equal to
			<unit2>, i.e. 'Between 3 days and 2 days ago' is transformed to
			'Between 3 and 2 days ago'. If the urls are empty, ' None' is
			appended to the header.
		"""
		first_unit = ''
		if time_range[1] != time_range[3]:
			first_unit = '{0} '.format(time_range[1])
		header = 'Between {0} {1}'.format(time_range[0], first_unit)
		header += 'and {0} {1} ago:'.format(time_range[2], time_range[3])

		return header if urls else header + ' None'

	def last_header(self, time_point, urls):
		"""
		Returns a header for an open-ended time-range.

		Arguments:
			time_point (tuple): The lower-bound time-point.
			urls (list|tuple): The urls, to check if empty.

		Returns:
			A header of the schema 'Last [<span>] <unit>'. The span is
			only shown if it is greater 1, such that 'Last 1 day' will be
			displayed as 'Last day'. If the urls are empty, ' None' is 
			appended to the header.
		"""
		span = '{0} '.format(time_point[0]) if time_point[0] > 1 else ''
		header = 'Last {0}{1}:'.format(span, time_point[1])

		return header if urls else header + ' None'

	def get_date(self, time_point, base=None):
		"""
		Returns a UNIX timestamp for a given time-point.

		The timestamp is base timestamp (either supplied to the function
		call or time.time()) minus the time_point, parsed into seconds.

		Arguments:
			time_point (tuple): A time_point tuple of the schema (span, unit).
			base (int): The base time-value from which to subtract the offset,
						defaults to time.time().

		Returns:
			A UNIX timestamp, in seconds, as an integer (!).
		"""
		span = time_point[0]
		unit = time_point[1]
		if unit.endswith('s'):
			unit = unit[:-1]
		offset = span * self.delta[unit]
		base = base or datetime.now()

		return base - offset

	def get_boundaries(self, time_range, base=None):
		"""
		Gets a lower and upper-bound timestamp for a given time_range.

		Also checks whether the starting point is before the end point,
		and throws a UsageError if not (the user passed the range in the
		wrong format, e.g. -r 5 day 7 weeks).

		Arguments:
			time_range (tuple): The time-range following the
								schema (span1, unit1, span2, unit2).
			base (int): The base timestamp to pass to get_date().

		Raises:
			errors.UsageError: if the starting point is after the end point.

		Returns:
			The begin and end timestamps (datetime objects).

		"""
		begin = self.get_date(time_range[:2], base)
		end = self.get_date(time_range[2:], base)
		if end < begin:
			what = "Illegal time range 'between {0} and {1} and {2} {3} "\
				   "ago' (start must precede end)"\
				   '!'.format(time_range[0], time_range[1],
							  time_range[2], time_range[3])
			raise lnk.errors.UsageError(what)
		return begin, end

	def request(self):
		"""
		Requests the link-history from the bit.ly API.

		Returns:
			A list of links.
		"""
		api = self.get_api()

		data = []
		response = {'nextPageToken': None}
		while 'nextPageToken' in response:
			request = api.list(start_token=response['nextPageToken'])
			response = self.execute(request, 'retrieve history')
			data += response['items']

		return data

	def listify(self, urls, limit, expanded, both, pretty):
		"""
		Returns a list of lines for a list of urls.

		Also performs limiting according to the 'limit' parameter.

		Arguments:
			urls (tuple|list): The urls to listify.
			limit (int): A limit to the number of urls to listify.
			expanded (bool): Whether to show expanded or short links.
			both (bool): Whether to show both short and expanded links.
			pretty (bool): Whether to format the output.

		Returns:
			A list of lines, ready for input. The length of the returned
			list will always be at most equal to the limit, if the limit
			is not None.
		"""
		lines = []
		for n, url in enumerate(urls):
			if n == limit:
				break
			line = self.lineify(url, expanded, both, pretty)
			lines.append(line)

		return lines

	def lineify(self, url, expanded, both, pretty):
		"""
		Returns an output-ready line for a given url.

		The line may include only the short url, only the long/expanded
		url or both, depending on the parameters passed.

		Arguments:
			url (str): The short url.
			expanded (bool): Whether or not to show the expanded link.
			both (bool): Whether or not to show both the short and the
						 expanded link (takes precedence over 'expanded').
			pretty (bool): Whether or not to make the output pretty.

		Returns:
			The short url if neither 'expanded' nor 'both' is set,
			both the short and the expanded url, in the schema '<short> =>
			<expanded>' if 'both' is set, else only the expanded link if
			'expanded' is set. If 'pretty' is True, the line is returned
			in a formatted fashion (with color), else in plain format.
		"""
		if both:
			line = '{0} => {1}'.format(url.short, url.long)
		elif expanded:
			line = url.long
		else:
			line = url.short

		if pretty:
			marked = self.list_item.format(line)
			line = ecstasy.beautify(marked, ecstasy.Color.Red)

		return line

	@staticmethod
	def filter(urls, begin, end):
		"""
		Filters a list of urls according to a lower and upper-bound time-point.

		Arguments:
			urls (list|tuple): A list/tuple of urls to filter.
			begin (int): The lower-bound datetime object (all links mus be
						 created at or after this point in time).
			end (int): The upper-bound datetime object (all links mus be
						 created at or before this point in time).

		Returns:
			A list of filtered urls.

		Note:
			The passed list of urls is not modified directly, i.e.
			the filtered list is a new list.
		"""
		filtered = []
		for url in urls:
			if url.created >= begin and url.created <= end:
				filtered.append(url)

		return filtered

	@staticmethod
	def process(data):
		"""
		Processes data retrieved from the API.

		Given a url, this method returns a History.Url object constisting
		of the short url, the expanded url and a datetime.datetime object
		for its date of creation.

		Arguments:
			data (list): The data from the HTTP request sent in the
						 request() method.

		Returns:
			A list of History.Url objects.
		"""
		urls = []
		for item in data:
			relevant = item['created'].split('.')[0]
			created = datetime.strptime(relevant, '%Y-%m-%dT%H:%M:%S')
			url = History.Url(item['id'], item['longUrl'], created)
			urls.append(url)

		return urls



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""The command-line interface to the bit.ly client."""

import click

import lnk.cli
import lnk.config
import lnk.errors

import lnk.tinyurl.link

lnk_config = lnk.config.Manager('lnk')['settings']
tinyurl_config = lnk.config.Manager('tinyurl')
link_config = tinyurl_config['commands']['link']

@click.group(invoke_without_command=True,
			 no_args_is_help=True,
			 context_settings=dict(ignore_unknown_options=True),
			 help='tinyurl command-line client.')
@click.option('-v',
			  '--verbose',
			  count=True,
			  help='Increments the level of verbosity.')
@click.option('-l',
			  '--level',
			  nargs=1,
			  type=int,
			  help='Controls the level of verbosity.')
@click.argument('args', nargs=-1)
@click.version_option(version=tinyurl_config['version'],
					  message='tinyurl API v%(version)s')
@click.pass_context
def main(context, verbose, level, args):
	"""
	Main entry-point to the tinyurl API from the command-line.

	All command-calls are handled here. Before passing on all command-specific
	arguments to the appropriate command, the verbosity setting is handled here,
	which is of use when catching any exceptions thrown by any commands.
	Moreover, this entry-point takes care of making the 'link' command of the
	service its default command, such that the user can type 'lnk ...' when
	meaning 'lnk tinyurl link ...'.
	"""
	verbosity = lnk.cli.get_verbosity(verbose, level)
	if args and args[0] == tinyurl_config['settings']['command']:
		args = args[1:] or ['--help']
	catch = lnk.errors.Catch(verbosity, link.get_help(context))
	catch.catch(link.main, args, standalone_mode=False)

@main.command()
@click.option('-c/-n',
			  '--copy/--no-copy',
			  default=lnk_config['copy'],
			  help='Whether or not to copy the first link to the clipboard.')
@click.option('-q/-l',
			  '--quiet/--loud',
			  default=False,
			  help='Whether or not to print the shortened links.')
@click.option('-s',
			  '--shorten',
			  multiple=True,
			  metavar='URL',
			  help='Shorten a long url.')
@click.option('--pretty/--plain',
			  default=link_config['settings']['pretty'],
			  help='Whether to show the links in a pretty box or as a plain list.')
@click.argument('urls', nargs=-1)
def link(copy, quiet, shorten, urls, pretty):
	"""Link shortening."""
	if not urls and not shorten:
		raise lnk.errors.UsageError('Please supply at least one URL.')
	lnk.tinyurl.link.echo(copy, quiet, shorten + urls, pretty)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""Information-retrieval for bitlink."""

from __future__ import unicode_literals

import click
import time

import lnk.abstract
import lnk.beauty

from lnk.bitly.command import Command

def echo(*args):
	"""
	Executes an info command and echoes its output.

	Arguments:
		args (variadic): The arguments to pass to an
						 Info instance's fetch() method.
	"""
	click.echo(Info().fetch(*args))

class Info(Command):
	"""
	Class to retrieve information about a bitlink.

	Information about a link may include when it was created, its expanded
	url, its tags, any note associated with it, when it was last modified
	and other things. This information is returned in a pretty list. The
	list can be put into a box for terminal output, or can be returned raw
	for internal use (such as by the stats command). Note that the information
	retrieved combines data from the /info and user/link_history endpoints.

	Note:
		A 'bitlink' is a link shortened with bit.ly.

	Attributes:
		raw (bool): Whether or not to return the output as a raw list,
					or as a string, in a box.
		sets (dict): A complete dictionary of the available sets of information.
		reverse (dict): A reverse mapping of the above-mentioned sets.
	"""
	def __init__(self, raw=False):
		super(Info, self).__init__('info')

		self.raw = raw
		self.sets = self.config['sets']
		# Dictionary comprehensions don't work for Python < 2.7
		self.reverse = dict((value, key) for (key, value) in self.sets.items())

	def fetch(self, only, hide, hide_empty, urls):
		"""
		Fetches the link information.

		Arguments:
			only (tuple): A tuple of strings representing the sets to include
						  in the response ('only' these will be included).
			hide (tuple): A tuple of strings representing the sets to hide
						  from the response (either from all possible sets
						  if only is empty, or else from those selected).
			hide_empty (bool): Whether or not to show things that are empty,
							   such as an empty list of tags.
			urls (tuple): A tuple of urls to fetch information for.

		Returns:
			A plain list of the raw lines if the 'raw' attribute is True,
			else a boxified, pretty string.
		"""
		sets = lnk.abstract.filter_sets(self.sets, only, hide)

		result = []
		threads = []
		for url in urls:
			self.queue.put(url.strip())
			t = self.new_thread(self.request, sets.values(), result, hide_empty)
			threads.append(t)
		self.join(threads)

		return result if self.raw else lnk.beauty.boxify(result)

	def request(self, sets, result, hide_empty):
		"""
		Requests all information about a url.

		The information returned for a URL is a combination of the data
		retrieved from the /info endpoint, and the data retrieved from the
		/user/link_history endpoint. Both these sets of information are
		retrieved here and combined into one set of information for a URL.
		This method is run in a separate thread for each URL, and as such
		uses a queue to get a URL and then appends the information to the
		list supplied in the argument. This is done in a thread-safe, locked
		way.

		Arguments:
			sets (dict): The sets of information to include in the data.
			result (list): The list of results for each URL, to which to
						   append the data requested.
			hide_empty(bool): Whether or not to hide empty things, passed on
							  to lineify.
		"""
		url = self.queue.get()
		data = self.request_info(url)
		data.update(self.request_history(url))

		selection = dict((key, data[key]) for key in data if key in sets)
		lines = self.lineify(url, selection, hide_empty)

		with self.lock:
			result.append(lines)

	def request_info(self, url):
		"""
		Requests information for a url.

		Requests the part of the information that can be fetched from the /info
		endpoint.

		Arguments:
			url (str): The bitlink to request information for.

		Returns:
			The information for the bitlink supplied.
		"""
		response = self.get(self.endpoints['info'], dict(shortUrl=url))
		response = self.verify(response,
							   "retrieve information for '{0}'".format(url),
							   'info')

		return response

	def request_history(self, url):
		"""
		Requests more information for a url.

		Requests the part of the information that can be fetched from the
		/user/link_history endpoint. This request requires oauth2 authorization.

		Arguments:
			url (str): The bitlink to request information for.

		Returns:
			The information for the bitlink supplied.
		"""
		response = self.get(self.endpoints['history'], dict(link=url))
		response = self.verify(response,
				   			   "retrieve additional "
				   			   "information for '{0}'".format(url),
				   			   'link_history')

		return response

	def lineify(self, url, data, hide_empty):
		"""
		Returns a list of lines for given (information) data.

		The lines returned include a header with the relevant URL,
		followed by string-formatted key/value pairs. If the value for
		a given key is a list, this list is further formatted and prettified.

		Arguments:
			url (str): The URL of which the data should be formatted.
			data (dict): The data (information) for the url.
			hide_empty (bool): Whether or not to hide empty things, such
							   as an empty list of tags.
		Returns:
			A list of lines, ready for output.
		"""
		lines = ['URL: {0}'.format(url)]
		for key, value in data.items():
			# 0 does not mean empty, while an empty
			# list/tuple/iterable and also string does
			if hide_empty and (not value and value != 0):
				continue
			if isinstance(value, list):
				lines.append(self.format(key, value))
				# A list item is of the schema ' + <thing>'
				# where the plus is in red color
				lines += [self.list_item.format(i) for i in value]
			else:
				lines.append(self.format(key, value))

		return lines

	def format(self, key, value):
		"""
		Handles formatting of a data-point.

		Handles cases in which the representation of a key or value is not
		directly suitable for output, such as when the value is a UNIX
		timestamp, which should really be formatted into a proper ctime
		representation. Also changes booleans into 'Yes'/'No'.

		Arguments:
			key (str): The key of the data-point.
			value (str): The value of the data-point.

		Returns:
			A formatted string of the schema '<key>: <value>'. What '<key>'
			and '<value>' end up being is highly-dependent on the actual
			values of the arguments.
		"""
		# The key is now what is returned from the bit.ly API, usually
		# something ugly such as 'created_at'. This must be
		# reverse-mapped to the string-representation used by lnk.
		key = self.reverse[key]
		if key == 'created' or key == 'modified':
			value = time.ctime(value)
		elif key == 'user' and value is None:
			value = 'Not public'
		elif key == 'privacy':
			# the set is '--only privacy', but
			# want to display 'Private: ...'
			key = 'private'

		if isinstance(value, bool):
			value = 'Yes' if value else 'No'
		elif not value:
			value = 'None'
		elif isinstance(value, list):
			value = ''

		return '{0}: {1}'.format(key.title(), value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Convenience path declarations."""

import os
import sys

TEST_PATH = os.path.dirname(os.path.abspath(__file__))
ROOT_PATH = os.path.dirname(TEST_PATH)
CONFIG_PATH = os.path.join(ROOT_PATH, 'config')

CREDENTIALS_PATH = os.path.join(TEST_PATH, 'googl', '.credentials')

sys.path.insert(0, ROOT_PATH)



#!/usr/bin/env python
#! -*- coding: utf-8 -*-

"""
Methods used to configure settings in the config files.

Note that these methods are not in any-way object-oriented or
general-purpose, they solely execute the config command.
"""

import click

import lnk.config


def echo(*args):
	"""
	Echoes the return value of configure() to stdout.

	Arguments:
		args (variadic): The arguments to forward to configure().
	"""
	click.echo(configure(*args), nl=False)


def configure(which, command, keys, values, quiet, all_keys):
	"""
	Does the actual configuration-management.

	Arguments:
		which (str): The configuration file for which to configure settings
					 (e.g. 'lnk' or 'bitly').
		command (str): Optionally, the command of the which for which
					   to configure settings (e.g. 'link').
		keys (tuple): The keys to show.
		values (tuple): The values to give the keys.
		quiet (bool): Whether to output anything or just do the
					 configuration quietly.
		all_keys (bool): Whether to show all keys.

	Returns:
		If quiet is True, nothing, else either a string-joined list of lines
		showing all updates, if values were supplied, else only the keys
		and if no keys were found (e.g. if the --all-keys was passed but the
		settings dictionary is empty) the string 'Nothing to see...\n'.
	"""
	with lnk.config.Manager(which, write=True) as manager:
		# If the command option is not None, get the settings for that
		# command, otherwise we'll use settings for the which itself
		if command:
			manager = manager['commands'][command]
		manager = manager['settings']
		keys = manager.keys() if (not keys or all_keys) else keys
		keys = list(keys)
		values = list(values)

		if quiet:
			be_quiet(manager, keys, values)
		else:
			lines = [update(manager, key, values) for key in keys]
			output = '\n'.join(lines) if lines else 'Nothing to see...'

			return '{0}\n'.format(output)


def be_quiet(manager, keys, values):
	"""
	Handles the case where the quiet flag was set.

	The effective result is the same as if the quiet flag was not set,
	but it is done more efficiently because nothing needs to be formatted here.

	Arguments:
		keys (tuple): The keys.
		values (tuple): The values.
	"""
	while values:
		key = keys.pop(0)
		manager[key] = values.pop(0)


def update(manager, key, values):
	"""
	Updates the configuration and returns a formatted string for the update.

	Arguments:
		manager (config.manager.Manager): The manager instance used for
										  interacting with the configuration
										  file.
		key (str): The key to configure.
		values (tuple): The tuple of values. If not None, the first one will be
						popped out of this list and regarded as the new value
						for the key.

	Returns:
		A formatted line, following the schema
		'<key>: <old_value> => <new_value>'.
	"""
	line = '{0}: {1}'.format(key, format_value(manager[key]))
	if values:
		manager[key] = values.pop(0)
		line += ' => {0}'.format(manager[key])

	return line


def format_value(value):
	"""
	Formats a value for output.

	Arguments:
		value (?): The value to format.

	Returns:
		If the value is a list, returns a comma-separated string of the items
		in the list. If the value is a dictionary, returns a comma-separated
		string-representation of the key-value-pairs. Else returns the value,
		cast to a string.
	"""
	if isinstance(value, list):
		return ', '.join(map(str, value))
	elif isinstance(value, dict):
		return ', '.join(['{0}: {1}'.format(k, v) for k, v in value.items()])
	return str(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import apiclient.discovery
import datetime
import googleapiclient.discovery
import httplib2
import oauth2client.file
import os
import pytest

from collections import namedtuple

import tests.paths

import lnk.googl.command
import lnk.errors

VERSION = 1
KEY = 'AIzaSyAoXKM_AMBafkXqmVeqJ82o9B9NPCTvXxc'
URL = 'https://www.googleapis.com/urlshortener'
API = apiclient.discovery.build('urlshortener',
								'v{0}'.format(VERSION),
								developerKey=KEY).url()


def get(url, projection=None):
	request = API.get(shortUrl=url, projection=projection)
	response = request.execute()

	return response

@pytest.fixture(scope='module')
def fixture():
	Fixture = namedtuple('Fixture', [
		'command',
		'url'
		])

	lnk.command = lnk.googl.command.Command('link', tests.paths.CREDENTIALS_PATH)

	print(10, os.listdir(os.path.dirname(tests.paths.CREDENTIALS_PATH)))
	url = 'http://goo.gl/Euc5'

	return Fixture(lnk.command, url)


def test_initializes_well(fixture):
	assert hasattr(fixture.command, 'credentials_path')
	assert hasattr(fixture.command, 'credentials')


def test_get_api_works(fixture):

	print(6, os.listdir(tests.paths.TEST_PATH))
	print(7, os.listdir(os.path.dirname(os.path.realpath(__file__))))
	print(8, oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH).get())
	print(9, oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH).get().access_token)

	result = fixture.command.get_api()

	assert isinstance(result, googleapiclient.discovery.Resource)
	assert all(hasattr(result, i) for i in ['get', 'insert', 'list'])


def test_execute_works_for_healthy_request(fixture):
	request = API.get(shortUrl=fixture.url)
	try:
		response = fixture.command.execute(request)
	except lnk.errors.HTTPError:
		pytest.fail('googl.command.execute threw '
					'an HTTPError for a healthy request!')

	assert isinstance(response, dict)


def test_execute_fails_for_bad_request(fixture):
	request = API.get(shortUrl='banana')

	with pytest.raises(lnk.errors.HTTPError):
		fixture.command.execute(request)


def test_execute_throws_error_with_correct_message(fixture):
	request = API.get(shortUrl='banana')

	with pytest.raises(lnk.errors.HTTPError) as error:
		fixture.command.execute(request, 'badness')

		assert error.value.what == 'badness'


def test_get_works(fixture):
	result = fixture.command.get(fixture.url)
	expected = get(fixture.url)

	assert isinstance(result, dict)
	assert result == expected


def test_authorize_works(fixture):
	result = fixture.command.authorize()

	assert isinstance(result, httplib2.Http)


def test_authorize_refreshes_well():
	storage = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	credentials = storage.get()
	old = datetime.datetime.now() - datetime.timedelta(days=100)
	credentials.token_expiry = old
	storage.put(credentials)

	command = lnk.googl.command.Command('link', tests.paths.CREDENTIALS_PATH)
	command.authorize()

	credentials = storage.get()

	assert credentials.token_expiry != old


def test_authorize_throws_if_no_credentials(fixture):
	storage = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	credentials = storage.get()
	os.remove(tests.paths.CREDENTIALS_PATH)

	with pytest.raises(lnk.errors.AuthorizationError):
		fixture.command.authorize()

	storage.put(credentials)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy
import ecstasy
import oauth2client.file
import pytest
import requests

from collections import namedtuple

import tests.paths
import lnk.googl.stats
import lnk.googl.info
import lnk.config

VERSION = 1
API = 'https://www.googleapis.com/urlshortener'
KEY = 'AIzaSyAoXKM_AMBafkXqmVeqJ82o9B9NPCTvXxc'

def request_stats(url):
	response = requests.get('{0}/v{1}/url'.format(API, VERSION),
							params=dict(projection='FULL',
										shortUrl=url,
										key=KEY))
	data = response.json()
	data['URL'] = url
	del data['kind']
	del data['id']

	return data

@pytest.fixture(scope='module')
def fixture(request):
	Fixture = namedtuple('Fixture', [
		'stats',
		'info',
		'category',
		'url',
		'timespans',
		'full',
		'analytics',
		'first_level',
		'second_level',
		'forever_data',
		'category_data',
		'timespans_data'
		])

	category = {'browsers': 'browsers'}
	url = 'http://goo.gl/3U9mIa'
	timespans = ['month', 'day']
	first_level = ecstasy.beautify(' <+> {0}', ecstasy.Color.Red)
	second_level = ecstasy.beautify('   <-> {0}: {1}', ecstasy.Color.Yellow)
	full = request_stats(url)
	analytics = full['analytics']
	forever_data = analytics['allTime']
	category_data = forever_data['browsers']
	timespans = ['month', 'week']
	timespans_data = [analytics[i] for i in timespans]

	with lnk.config.Manager('googl', write=True) as manager:
		settings = manager['commands']['stats']['settings']
		old = settings['timespan']

	stats = lnk.googl.stats.Stats(raw=True)
	stats.credentials = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)
	info = lnk.googl.info.Info(raw=True)
	info.credentials = oauth2client.file.Storage(tests.paths.CREDENTIALS_PATH)

	def finalize():
		with lnk.config.Manager('googl', write=True) as manager:
			settings = manager['commands']['stats']['settings']
			settings['span'] = old

	request.addfinalizer(finalize)

	return Fixture(stats,
				   info,
				   category,
				   url,
				   timespans,
				   full,
				   analytics,
				   first_level,
				   second_level,
				   forever_data,
				   category_data,
				   timespans_data)


def test_format_makes_lines_pretty(fixture):
	result = fixture.stats.format('foo', 'bar')
	expected = 'Foo: bar'

	assert result == expected


def test_format_handles_special_keys_wells(fixture):
	result = fixture.stats.format('shortUrlClicks', 'foo')
	expected = 'Clicks: foo'

	assert result == expected

	result = fixture.stats.format('longUrl', 'foo')
	expected = 'Expanded: foo'

	assert result == expected


def test_get_timespans_removes_duplicates(fixture):
	things = [1, 4, 2, 3, 4, 1, 3, 3]
	result = fixture.stats.get_timespans(things, False)
	expected = set(things)

	assert result == expected


def test_get_timespans_picks_default_timespan_if_no_times():
	with lnk.config.Manager('googl', write=True) as manager:
		settings = manager['commands']['stats']['settings']
		settings['timespan'] = 'day'
	stats = lnk.googl.stats.Stats(raw=True)
	result = stats.get_timespans([], False)

	assert result == set(['day'])

def test_get_timespans_handles_default_forever_well():
	with lnk.config.Manager('googl', write=True) as manager:
		settings = manager['commands']['stats']['settings']
		settings['timespan'] = 'forever'
	stats = lnk.googl.stats.Stats(raw=True)
	result = stats.get_timespans([], False)

	assert result == set(['allTime'])


def test_get_timespans_handles_default_two_hours_well():
	with lnk.config.Manager('googl', write=True) as manager:
		settings = manager['commands']['stats']['settings']
		settings['timespan'] = 'two-hours'
	stats = lnk.googl.stats.Stats(raw=True)
	result = stats.get_timespans([], False)

	assert result == set(['twoHours'])

def test_get_timespans_handles_forever_well(fixture):
	result = fixture.stats.get_timespans([], True)

	assert result == set(['allTime'])


def test_get_timespans_handles_two_hours_well(fixture):
	result = fixture.stats.get_timespans(['two-hours'], False)

	assert result == set(['twoHours'])


def test_get_timespans_works(fixture):
	result = fixture.stats.get_timespans(fixture.timespans, False)

	assert result == set(fixture.timespans)


def test_sub_listify_works_in_normal_cases(fixture):
	result = fixture.stats.sub_listify(fixture.category,
										 fixture.category_data,
										 None,
										 False)
	expected = []
	for point in fixture.category_data:
		line = fixture.second_level.format(point['id'], point['count'])
		expected.append(line)

	assert result == expected


def test_sub_listify_limits_well(fixture):
	result = fixture.stats.sub_listify(fixture.category,
										fixture.category_data,
										1,
										False)
	expected = [fixture.second_level.format(fixture.category_data[0]['id'],
											fixture.category_data[0]['count'])]
	assert len(result) == 1
	assert result == expected


def test_sub_listify_handles_unknown_well(fixture):
	data = [dict(id='unknown', count=123)]
	result = fixture.stats.sub_listify(fixture.category,
									   data,
									   None,
									   False)
	expected = [fixture.second_level.format('Unknown', 123)]

	assert result == expected


def test_sub_listify_leaves_countries_short(fixture):
	data = [dict(id='DE', count=123)]
	result = fixture.stats.sub_listify('countries',
									   data,
									   None,
									   False)
	expected = [fixture.second_level.format('DE', 123)]

	assert result == expected

def test_sub_listify_expands_countries(fixture):
	data = [dict(id='DE', count=123)]
	result = fixture.stats.sub_listify('countries',
									   data,
									   None,
									   True)
	expected = [fixture.second_level.format('Germany', 123)]

	assert result == expected


def test_get_header_handles_forever_well(fixture):
	result = fixture.stats.get_header('allTime')
	expected = fixture.first_level.format('Since forever:')

	assert result == expected


def test_get_header_handles_two_hours_well(fixture):
	result = fixture.stats.get_header('twoHours')
	expected = fixture.first_level.format('Last two hours:')

	assert result == expected


def test_get_header_handles_normal_cases_well(fixture):
	result = fixture.stats.get_header('month')
	expected = fixture.first_level.format('Last month:')

	assert result == expected


def test_request_format_is_correct(fixture):
	fixture.stats.queue.put(fixture.url)
	result = fixture.stats.request(False)

	assert isinstance(result, dict)
	assert 'URL' in result
	assert (isinstance(result['URL'], str)
		 or isinstance(result['URL'], unicode))
	assert 'analytics' in result
	assert isinstance(result['analytics'], dict)


def test_requests_well_with_info(fixture):
	fixture.stats.queue.put(fixture.url)
	result = fixture.stats.request(True)

	assert result == fixture.full


def test_requests_well_without_info(fixture):
	fixture.stats.queue.put(fixture.url)
	result = fixture.stats.request(False)
	expected = fixture.full.copy()
	for i in ['status', 'created', 'longUrl']:
		del expected[i]

	assert result == expected


def test_listify_formats_timespans_well(fixture):
	timespans = ['allTime'] + fixture.timespans
	result = fixture.stats.listify(fixture.analytics,
								   fixture.category,
								   timespans,
								   False,
								   None)

	expected = [fixture.first_level.format('Since forever:')]
	for i in fixture.timespans:
		header = fixture.first_level.format('Last {0}:'.format(i))
		expected.append(header)

	for i in expected:
		assert i in result


def test_listify_filters_timespans_well(fixture):
	result = fixture.stats.listify(fixture.analytics,
								   fixture.category,
								   fixture.timespans,
								   False,
								   None)
	for i in fixture.timespans:
		line = 'Last {0}:'.format(i)
		formatted = fixture.first_level.format(line)
		assert formatted in result

	unwanted = ['Since forever', 'Last day:', 'Last two hours:']
	for i in unwanted:
		formatted = fixture.first_level.format(i)
		assert formatted not in result


def test_listify_handles_clicks_well(fixture):
	result = fixture.stats.listify(fixture.analytics,
								   {'clicks': 'shortUrlClicks'},
								   [fixture.timespans[0]],
								   False,
								   None)
	clicks = fixture.analytics[fixture.timespans[0]]['shortUrlClicks']
	line = 'Last {0}: {1}'.format(fixture.timespans[0], clicks)
	expected = fixture.first_level.format(line)

	assert expected in result


def test_lineify_formats_headers_well(fixture):
	data = copy.deepcopy(fixture.full)
	result = fixture.stats.lineify(data,
								   fixture.category,
								   fixture.timespans,
								   False,
								   None)
	for key, value in fixture.full.items():
		if key != 'analytics':
			expected = fixture.stats.format(key, value)
			assert expected in result

def test_get_stats_works(fixture):
	"""
	result = []
	fixture.stats.get_stats(result,
							fixture.category,
							fixture.timespans,
							True,
							False,
							None)
	assert result[0] == fixture.full
	"""
	pass


def test_fetch_works_for_single_url(fixture):
	result = fixture.stats.fetch(list(fixture.category.keys()),
								 [],
								 fixture.timespans,
								 False,
								 None,
								 True,
								 False,
								 [fixture.url])
	data = copy.deepcopy(fixture.full)
	expected = fixture.stats.lineify(data,
									 fixture.category,
									 fixture.timespans,
									 False,
									 None)

	assert sorted(result[0]) == sorted(expected)


def test_fetch_limits_well(fixture):
	result = fixture.stats.fetch(list(fixture.category.keys()),
								 [],
								 fixture.timespans,
								 False,
								 1,
								 True,
								 False,
								 [fixture.url])
	data = copy.deepcopy(fixture.full)
	expected = fixture.stats.lineify(data,
									 fixture.category,
									 fixture.timespans,
									 False,
									 1)

	assert sorted(result[0]) == sorted(expected)

def test_fetch_works_for_many_urls(fixture):
	other = 'http://goo.gl/XBzv0g'
	result = fixture.stats.fetch(list(fixture.category.keys()),
								 [],
								 fixture.timespans,
								 False,
								 None,
								 True,
								 False,
								 [fixture.url, other])
	result = sorted(sorted(i) for i in result)
	data = copy.deepcopy(fixture.full)
	expected = []
	first = fixture.stats.lineify(data,
			  					   fixture.category,
								   fixture.timespans,
								   False,
								   None)
	expected.append(sorted(first))
	other_data = request_stats(other)
	second = fixture.stats.lineify(other_data,
			  					   fixture.category,
								   fixture.timespans,
								   False,
								   None)
	expected.append(sorted(second))

	assert result == sorted(expected)


def test_fetch_adds_info_well(fixture):
	result = fixture.stats.fetch(list(fixture.category.keys()),
								 [],
								 fixture.timespans,
								 False,
								 1,
								 True,
								 False,
								 [fixture.url])
	data = copy.deepcopy(fixture.full)
	expected = fixture.stats.lineify(data,
									 fixture.category,
									 fixture.timespans,
									 False,
									 1)

	assert sorted(result[0]) == sorted(expected)
